# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from uuid import UUID

import sqlalchemy.orm as orm
from sqlalchemy.orm import Mapped, relationship

from gen_epix.casedb.domain import DOMAIN, enum, model
from gen_epix.commondb.repositories.sa_model import (
    RowMetadataMixin,
    create_mapped_column,
    create_table_args,
)

Base: type = orm.declarative_base(name=enum.ServiceType.GEO.value)


class RegionSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.RegionSet)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.RegionSet, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.RegionSet, "name")
    region_code_as_label: Mapped[bool] = create_mapped_column(
        DOMAIN, model.RegionSet, "region_code_as_label"
    )


class RegionSetShape(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.RegionSetShape)

    region_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.RegionSetShape, "region_set_id"
    )
    scale: Mapped[float] = create_mapped_column(DOMAIN, model.RegionSetShape, "scale")
    geo_json: Mapped[str] = create_mapped_column(
        DOMAIN, model.RegionSetShape, "geo_json"
    )

    region_set: Mapped[model.RegionSet] = relationship(
        RegionSet, foreign_keys=[region_set_id]
    )


class Region(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Region)

    region_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Region, "region_set_id"
    )
    code: Mapped[str] = create_mapped_column(DOMAIN, model.Region, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.Region, "name")
    centroid_lat: Mapped[float] = create_mapped_column(
        DOMAIN, model.Region, "centroid_lat"
    )
    centroid_lon: Mapped[float] = create_mapped_column(
        DOMAIN, model.Region, "centroid_lon"
    )
    center_lat: Mapped[float] = create_mapped_column(DOMAIN, model.Region, "center_lat")
    center_lon: Mapped[float] = create_mapped_column(DOMAIN, model.Region, "center_lon")

    region_set: Mapped[model.RegionSet] = relationship(
        RegionSet, foreign_keys=[region_set_id]
    )


class RegionRelation(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.RegionRelation)

    from_region_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.RegionRelation, "from_region_id"
    )
    to_region_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.RegionRelation, "to_region_id"
    )
    relation: Mapped[enum.RegionRelationType] = create_mapped_column(
        DOMAIN, model.RegionRelation, "relation"
    )

    from_region: Mapped[model.Region] = relationship(
        Region, foreign_keys=[from_region_id]
    )
    to_region: Mapped[model.Region] = relationship(Region, foreign_keys=[to_region_id])



import ssl
from collections.abc import Hashable
from pathlib import Path

LOCAL_HOSTS = {"localhost", "127.0.0.1", "0.0.0.0"}


def serialize_id(value: Hashable) -> str | None:
    return str(value) if value else None


def create_ssl_context(
    host: str, ssl_cert_file: Path | str | None = None, disable_ssl_verification: bool = False
) -> ssl.SSLContext | bool:
    """Get SSL verification setting using similar logic to OidcClient."""
    if disable_ssl_verification:
        return False

    # For local development hosts, disable verification
    if host in LOCAL_HOSTS:
        return False

    # If a cert file is given, use that for verification
    if ssl_cert_file is not None:
        if isinstance(ssl_cert_file, str):
            ssl_cert_file = Path(ssl_cert_file)
        ssl_context = ssl.create_default_context()
        ssl_context.load_verify_locations(ssl_cert_file.absolute().as_posix())
        return ssl_context

    # Use default verification
    return True



from typing import ClassVar
from uuid import UUID

import gen_epix.casedb.domain.model.ontology as model
from gen_epix.casedb.domain import enum
from gen_epix.commondb.domain.command import CrudCommand, UpdateAssociationCommand

# Non-CRUD


class DiseaseEtiologicalAgentUpdateAssociationCommand(UpdateAssociationCommand):
    ASSOCIATION_CLASS: ClassVar = model.Etiology
    LINK_FIELD_NAME1: ClassVar = "disease_id"
    LINK_FIELD_NAME2: ClassVar = "etiological_agent_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.Etiology]


# CRUD


class ConceptCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Concept


class ConceptSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptSet


class ConceptRelationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptRelation


class DiseaseCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Disease


class EtiologicalAgentCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.EtiologicalAgent


class EtiologyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Etiology



import logging
import uuid
from typing import Any

from fastapi import Request
from fastapi.security.utils import get_authorization_scheme_param
import jwt

from gen_epix.fastapp import exc
from gen_epix.fastapp.log import BaseLogItem, LogItem
from gen_epix.fastapp.services.auth.idp_client import IdpClient
from gen_epix.fastapp.services.auth.model import Claims, IdentityProvider


class MockIDPClient(IdpClient):

    def __init__(
        self,
        logger: logging.Logger | None = None,
        log_item_class: type[BaseLogItem] = LogItem,
        **kwargs: Any,
    ):
        self._id: uuid.UUID = kwargs.get("id", uuid.uuid4())  # type: ignore[assignment]
        # Set input properties and initialise some
        self._logger = logger
        self._log_item_class = log_item_class

    @property
    def id(self) -> uuid.UUID:
        return self._id

    def get_identity_provider(self) -> IdentityProvider:
        raise NotImplementedError()

    async def get_claims_from_jwt(
        self, jwt_token: str
    ) -> dict[str, str | int | bool | list[str]] | None:
        raise NotImplementedError()

    def get_claims_from_userinfo(
        self, access_token: str
    ) -> dict[str, str | int | bool | list[str]]:
        raise NotImplementedError()

    async def __call__(self, request: Request) -> Claims | None:
        if authorization := request.headers.get("authorization"):
            scheme, token = get_authorization_scheme_param(authorization)
            if scheme.upper() == "BEARER":
                # TODO: check if this is a security risk
                # or whether it should return an error
                try:
                    claims = jwt.decode(token, options={"verify_signature": False})
                    if not claims:
                        return None
                    return Claims(
                        claims=claims, scheme=scheme, token=token, idp_client_id=self.id
                    )
                except exc.AuthException as exception:
                    if self._logger:
                        self._logger.warning(
                            self._log_item_class(
                                code="e86a3bd6",  # type: ignore[arg-type]
                                exception=exception,  # type: ignore[arg-type]
                            ).dumps()
                        )
                    return None

            else:
                # Authorization scheme not implemented
                if self._logger:
                    self._logger.warning(
                        self._log_item_class(
                            code="dec5fffe",  # type: ignore[arg-type]
                            msg=f"Authorization scheme {scheme} not implemented",  # type: ignore[arg-type]
                        ).dumps()
                    )
                return None

        if self._logger:
            self._logger.warning(
                self._log_item_class(
                    code="e14344c3",  # type: ignore[arg-type]
                    msg="No authorisation information provided in header",  # type: ignore[arg-type]
                ).dumps()
            )
        return None



# pylint: disable=too-few-public-methods

from uuid import UUID

import sqlalchemy.orm as orm
from sqlalchemy.orm import Mapped

from gen_epix.commondb.repositories.sa_model import (
    RowMetadataMixin,
    create_mapped_column,
    create_table_args,
)
from gen_epix.seqdb.domain import DOMAIN, enum, model
from gen_epix.seqdb.repositories.sa_model.base import (
    AlignmentMixin,
    CodeMixin,
    ProtocolMixin,
    QualityMixin,
    SeqMixin,
)

Base: type = orm.declarative_base(name=enum.ServiceType.SEQ.value)

# TODO: add SA relationship calls


class AlignmentProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.AlignmentProtocol)

    is_multiple: Mapped[bool] = create_mapped_column(
        DOMAIN, model.AlignmentProtocol, "is_multiple"
    )


class AssemblyProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.AssemblyProtocol)

    has_manual_curation: Mapped[bool] = create_mapped_column(
        DOMAIN, model.AssemblyProtocol, "has_manual_curation"
    )


class AstProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.AstProtocol)

    antimicrobial_names: Mapped[list[str]] = create_mapped_column(
        DOMAIN, model.AstProtocol, "antimicrobial_names"
    )
    is_predicted: Mapped[bool] = create_mapped_column(
        DOMAIN, model.AstProtocol, "is_predicted"
    )


class RefSnp(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.RefSnp)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.RefSnp, "code")
    ref_seq_id: Mapped[str] = create_mapped_column(DOMAIN, model.RefSnp, "ref_seq_id")
    position: Mapped[str] = create_mapped_column(DOMAIN, model.RefSnp, "position")
    nucleotide: Mapped[str] = create_mapped_column(DOMAIN, model.RefSnp, "nucleotide")


class SnpDetectionProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.SnpDetectionProtocol)


class KmerDetectionProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.KmerDetectionProtocol)


class RefSnpSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.RefSnpSet)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.RefSnpSet, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.RefSnpSet, "name")


class RefSnpSetMember(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.RefSnpSetMember)

    ref_snp_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.RefSnpSetMember, "ref_snp_set_id"
    )
    ref_snp_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.RefSnpSetMember, "ref_snp_id"
    )
    index: Mapped[int] = create_mapped_column(DOMAIN, model.RefSnpSetMember, "index")


class LibraryPrepProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.LibraryPrepProtocol)


class Locus(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Locus)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.Locus, "code")
    gene_code: Mapped[str] = create_mapped_column(DOMAIN, model.Locus, "gene_code")
    product_name: Mapped[str] = create_mapped_column(
        DOMAIN, model.Locus, "product_name"
    )


class LocusDetectionProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.LocusDetectionProtocol)


class LocusSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.LocusSet)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.LocusSet, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.LocusSet, "name")
    n_loci: Mapped[int] = create_mapped_column(DOMAIN, model.LocusSet, "n_loci")
    locus_ids: Mapped[list[UUID]] = create_mapped_column(
        DOMAIN, model.LocusSet, "locus_ids"
    )


class LocusSetMember(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.LocusSetMember)

    locus_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.LocusSetMember, "locus_set_id"
    )
    locus_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.LocusSetMember, "locus_id"
    )
    index: Mapped[int] = create_mapped_column(DOMAIN, model.LocusSetMember, "index")


class RefAllele(Base, RowMetadataMixin, SeqMixin):
    __tablename__, __table_args__ = create_table_args(model.RefAllele)

    locus_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.RefAllele, "locus_id")
    index: Mapped[int] = create_mapped_column(DOMAIN, model.RefAllele, "index")


class PcrProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.PcrProtocol)

    target_names: Mapped[list[str]] = create_mapped_column(
        DOMAIN, model.PcrProtocol, "target_names"
    )


class SeqCategorySet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SeqCategorySet)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.SeqCategory, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.SeqCategory, "name")


class SeqCategory(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SeqCategory)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.SeqCategory, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.SeqCategory, "name")
    seq_category_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqCategory, "seq_category_set_id"
    )


class SeqClassificationProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.SeqClassificationProtocol)

    is_taxonomic: Mapped[bool] = create_mapped_column(
        DOMAIN, model.SeqClassificationProtocol, "is_taxonomic"
    )


class SeqDistanceProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.SeqDistanceProtocol)

    max_stored_distance: Mapped[float] = create_mapped_column(
        DOMAIN, model.SeqDistanceProtocol, "max_stored_distance"
    )
    min_scale_unit: Mapped[float] = create_mapped_column(
        DOMAIN, model.SeqDistanceProtocol, "min_scale_unit"
    )
    seq_distance_protocol_type: Mapped[enum.SeqDistanceProtocolType] = (
        create_mapped_column(
            DOMAIN, model.SeqDistanceProtocol, "seq_distance_protocol_type"
        )
    )
    locus_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqDistanceProtocol, "locus_set_id"
    )
    ref_seq_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqDistanceProtocol, "ref_seq_id"
    )


class SubtypingScheme(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SubtypingScheme)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.SubtypingScheme, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.SubtypingScheme, "name")
    description: Mapped[str] = create_mapped_column(
        DOMAIN, model.SubtypingScheme, "description"
    )


class Taxon(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Taxon)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.Taxon, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.Taxon, "name")
    rank: Mapped[str] = create_mapped_column(DOMAIN, model.Taxon, "rank")
    ncbi_taxid: Mapped[int] = create_mapped_column(DOMAIN, model.Taxon, "ncbi_taxid")
    ictv_ictv_id: Mapped[str] = create_mapped_column(
        DOMAIN, model.Taxon, "ictv_ictv_id"
    )
    snomed_sctid: Mapped[int] = create_mapped_column(
        DOMAIN, model.Taxon, "snomed_sctid"
    )
    subtyping_scheme_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Taxon, "subtyping_scheme_id"
    )
    ncbi_ancestor_taxids: Mapped[list[int]] = create_mapped_column(
        DOMAIN, model.Taxon, "ncbi_ancestor_taxids"
    )
    ancestor_taxon_ids: Mapped[list[UUID]] = create_mapped_column(
        DOMAIN, model.Taxon, "ancestor_taxon_ids"
    )


class TaxonLocusLink(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.TaxonLocusLink)

    taxon_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.TaxonLocusLink, "taxon_id"
    )
    locus_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.TaxonLocusLink, "locus_id"
    )


class TaxonomyProtocol(Base, RowMetadataMixin, ProtocolMixin):
    __tablename__, __table_args__ = create_table_args(model.TaxonomyProtocol)


class TaxonSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.TaxonSet)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.TaxonSet, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.TaxonSet, "name")


class TaxonSetMember(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.TaxonSetMember)

    taxon_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.TaxonSetMember, "taxon_set_id"
    )
    taxon_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.TaxonSetMember, "taxon_id"
    )


class TreeAlgorithm(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.TreeAlgorithm)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.TreeAlgorithm, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.TreeAlgorithm, "name")
    description: Mapped[str] = create_mapped_column(
        DOMAIN, model.TreeAlgorithm, "description"
    )
    tree_algorithm_class_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.TreeAlgorithm, "tree_algorithm_class_id"
    )
    is_ultrametric: Mapped[bool] = create_mapped_column(
        DOMAIN, model.TreeAlgorithm, "is_ultrametric"
    )
    rank: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.TreeAlgorithmClass, "rank"
    )


class TreeAlgorithmClass(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.TreeAlgorithmClass)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.TreeAlgorithmClass, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.TreeAlgorithmClass, "name")
    is_seq_based: Mapped[bool] = create_mapped_column(
        DOMAIN, model.TreeAlgorithmClass, "is_seq_based"
    )
    is_dist_based: Mapped[bool] = create_mapped_column(
        DOMAIN, model.TreeAlgorithmClass, "is_dist_based"
    )
    rank: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.TreeAlgorithmClass, "rank"
    )


class Allele(Base, RowMetadataMixin, SeqMixin, QualityMixin):
    __tablename__, __table_args__ = create_table_args(model.Allele)

    locus_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Allele, "locus_id")


class AlleleAlignment(Base, RowMetadataMixin, AlignmentMixin, QualityMixin):
    __tablename__, __table_args__ = create_table_args(model.AlleleAlignment)

    ref_allele_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.AlleleAlignment, "ref_allele_id"
    )
    allele_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.AlleleAlignment, "allele_id"
    )
    alignment_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.AlleleAlignment, "alignment_protocol_id"
    )


class AlleleProfile(Base, RowMetadataMixin, QualityMixin):
    __tablename__, __table_args__ = create_table_args(model.AlleleProfile)

    seq_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.AlleleProfile, "seq_id")
    locus_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.AlleleProfile, "locus_set_id"
    )
    locus_detection_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.AlleleProfile, "locus_detection_protocol_id"
    )
    n_loci: Mapped[int] = create_mapped_column(DOMAIN, model.AlleleProfile, "n_loci")
    allele_profile: Mapped[str] = create_mapped_column(
        DOMAIN, model.AlleleProfile, "allele_profile"
    )
    allele_profile_format: Mapped[str] = create_mapped_column(
        DOMAIN, model.AlleleProfile, "allele_profile_format"
    )
    allele_profile_hash_sha256: Mapped[bytes] = create_mapped_column(
        DOMAIN, model.AlleleProfile, "allele_profile_hash_sha256"
    )


class SnpProfile(Base, RowMetadataMixin, QualityMixin):
    __tablename__, __table_args__ = create_table_args(model.SnpProfile)

    ref_seq_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SnpProfile, "ref_seq_id"
    )
    seq_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.SnpProfile, "seq_id")
    snp_detection_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SnpProfile, "snp_detection_protocol_id"
    )
    snp_profile: Mapped[str] = create_mapped_column(
        DOMAIN, model.SnpProfile, "snp_profile"
    )
    snp_profile_format: Mapped[str] = create_mapped_column(
        DOMAIN, model.SnpProfile, "snp_profile_format"
    )
    snp_profile_hash_sha256: Mapped[bytes] = create_mapped_column(
        DOMAIN, model.SnpProfile, "snp_profile_hash_sha256"
    )


class KmerProfile(Base, RowMetadataMixin, QualityMixin):
    __tablename__, __table_args__ = create_table_args(model.KmerProfile)

    seq_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.KmerProfile, "seq_id")
    kmer_detection_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.KmerProfile, "kmer_detection_protocol_id"
    )
    kmer_profile: Mapped[str] = create_mapped_column(
        DOMAIN, model.KmerProfile, "kmer_profile"
    )
    kmer_profile_format: Mapped[str] = create_mapped_column(
        DOMAIN, model.KmerProfile, "kmer_profile_format"
    )
    kmer_profile_hash_sha256: Mapped[bytes] = create_mapped_column(
        DOMAIN, model.KmerProfile, "kmer_profile_hash_sha256"
    )


class AstMeasurement(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.AstMeasurement)

    sample_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.AstMeasurement, "sample_id"
    )
    ast_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.AstMeasurement, "ast_protocol_id"
    )
    ast_result: Mapped[str] = create_mapped_column(
        DOMAIN, model.AstMeasurement, "ast_result"
    )
    ast_result_format: Mapped[str] = create_mapped_column(
        DOMAIN, model.AstMeasurement, "ast_result_format"
    )
    index: Mapped[int] = create_mapped_column(DOMAIN, model.AstMeasurement, "index")


class AstPrediction(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.AstPrediction)

    seq_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.AstPrediction, "seq_id")
    ast_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.AstPrediction, "ast_protocol_id"
    )
    ast_result: Mapped[str] = create_mapped_column(
        DOMAIN, model.AstPrediction, "ast_result"
    )
    ast_result_format: Mapped[str] = create_mapped_column(
        DOMAIN, model.AstPrediction, "ast_result_format"
    )


class PcrMeasurement(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.PcrMeasurement)

    sample_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PcrMeasurement, "sample_id"
    )
    pcr_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PcrMeasurement, "pcr_protocol_id"
    )
    pcr_result: Mapped[str] = create_mapped_column(
        DOMAIN, model.PcrMeasurement, "pcr_result"
    )
    pcr_result_format: Mapped[str] = create_mapped_column(
        DOMAIN, model.PcrMeasurement, "pcr_result_format"
    )
    index: Mapped[int] = create_mapped_column(DOMAIN, model.PcrMeasurement, "index")


class ReadSet(Base, RowMetadataMixin, CodeMixin, QualityMixin):
    __tablename__, __table_args__ = create_table_args(model.ReadSet)

    fwd_uri: Mapped[str] = create_mapped_column(DOMAIN, model.ReadSet, "fwd_uri")
    rev_uri: Mapped[str] = create_mapped_column(DOMAIN, model.ReadSet, "rev_uri")
    fwd_file_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ReadSet, "fwd_file_id"
    )
    rev_file_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ReadSet, "rev_file_id"
    )
    fwd_reads_hash_sha256: Mapped[bytes] = create_mapped_column(
        DOMAIN, model.ReadSet, "fwd_reads_hash_sha256"
    )
    rev_reads_hash_sha256: Mapped[bytes] = create_mapped_column(
        DOMAIN, model.ReadSet, "rev_reads_hash_sha256"
    )
    library_prep_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ReadSet, "library_prep_protocol_id"
    )
    sequencing_run_code: Mapped[str] = create_mapped_column(
        DOMAIN, model.ReadSet, "sequencing_run_code"
    )


class RawSeq(Base, RowMetadataMixin, SeqMixin):
    __tablename__, __table_args__ = create_table_args(model.RawSeq)


class RefSeq(Base, RowMetadataMixin, SeqMixin):
    __tablename__, __table_args__ = create_table_args(model.RefSeq)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.RefSeq, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.RefSeq, "name")
    description: Mapped[str] = create_mapped_column(DOMAIN, model.RefSeq, "description")
    taxon_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.RefSeq, "taxon_id")
    genbank_accession_code: Mapped[str] = create_mapped_column(
        DOMAIN, model.RefSeq, "genbank_accession_code"
    )


class Sample(Base, RowMetadataMixin, CodeMixin):
    __tablename__, __table_args__ = create_table_args(model.Sample)

    props: Mapped[dict[str, str]] = create_mapped_column(DOMAIN, model.Sample, "props")


class Seq(Base, RowMetadataMixin, CodeMixin, QualityMixin):
    __tablename__, __table_args__ = create_table_args(model.Seq)

    sample_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Seq, "sample_id")
    read_set_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Seq, "read_set_id")
    read_set2_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Seq, "read_set2_id")
    assembly_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Seq, "assembly_protocol_id"
    )
    raw_seq_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Seq, "raw_seq_id")
    file_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Seq, "file_id")


class SeqAlignment(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SeqAlignment)

    seq_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.SeqAlignment, "seq_id")
    alignment_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqAlignment, "alignment_protocol_id"
    )
    contig_alignments: Mapped[list[model.ContigAlignment]] = create_mapped_column(
        DOMAIN, model.SeqAlignment, "contig_alignments"
    )


class SeqClassification(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SeqClassification)

    seq_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqClassification, "seq_id"
    )
    seq_classification_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqClassification, "seq_classification_protocol_id"
    )
    primary_category_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqClassification, "primary_category_id"
    )
    classification: Mapped[str] = create_mapped_column(
        DOMAIN, model.SeqClassification, "classification"
    )
    classification_format: Mapped[str] = create_mapped_column(
        DOMAIN, model.SeqClassification, "classification_format"
    )
    classification_hash_sha256: Mapped[bytes] = create_mapped_column(
        DOMAIN, model.SeqClassification, "classification_hash_sha256"
    )


class SeqDistance(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SeqDistance)

    seq_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.SeqDistance, "seq_id")
    seq_distance_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqDistance, "seq_distance_protocol_id"
    )
    allele_profile_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqDistance, "allele_profile_id"
    )
    snp_profile_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqDistance, "snp_profile_id"
    )
    kmer_profile_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqDistance, "kmer_profile_id"
    )
    distance_format: Mapped[str] = create_mapped_column(
        DOMAIN, model.SeqDistance, "distance_format"
    )
    distances: Mapped[str] = create_mapped_column(
        DOMAIN, model.SeqDistance, "distances"
    )


class SeqTaxonomy(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SeqTaxonomy)

    seq_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.SeqTaxonomy, "seq_id")
    taxonomy_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqTaxonomy, "taxonomy_protocol_id"
    )
    primary_taxon_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SeqTaxonomy, "primary_taxon_id"
    )
    taxonomy: Mapped[str] = create_mapped_column(DOMAIN, model.SeqTaxonomy, "taxonomy")
    taxonomy_format: Mapped[str] = create_mapped_column(
        DOMAIN, model.SeqTaxonomy, "taxonomy_format"
    )
    taxonomy_hash_sha256: Mapped[bytes] = create_mapped_column(
        DOMAIN, model.SeqTaxonomy, "taxonomy_hash_sha256"
    )



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import ClassVar, Self
from uuid import UUID

from pydantic import Field, model_validator

from gen_epix.commondb.domain.command import Command, CrudCommand
from gen_epix.seqdb.domain import enum, model

# Non-CRUD commands


class RetrieveCompleteContigCommand(Command):
    pass


class RetrieveCompleteAlleleProfileCommand(Command):
    pass


class RetrieveCompleteSnpProfileCommand(Command):
    pass


class RetrieveCompleteSeqCommand(Command):

    seq_ids: list[UUID]


class RetrieveCompleteSampleCommand(Command):
    pass


class RetrievePhylogeneticTreeCommand(Command):

    seq_distance_protocol_id: UUID
    tree_algorithm: enum.TreeAlgorithm
    seq_ids: list[UUID]
    leaf_names: list[str] | None

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if self.leaf_names is not None and len(self.leaf_names) != len(self.seq_ids):
            raise ValueError(
                "leaf_codes must be None or have the same length as seq_ids"
            )
        return self


class RetrieveSeqFastaCommand(Command):

    seq_ids: list[UUID] = Field(
        description="List of sequence IDs to retrieve in FASTA format.",
    )
    wrap: int | None = Field(
        default=80,
        description="Number of characters to wrap the sequence lines.",
    )


class RetrieveMultipleAlignmentCommand(Command):
    pass


class GenerateMultipleAlignmentCommand(Command):
    pass


class GeneratePhylogeneticTreeCommand(Command):
    pass


# CRUD commands


class TreeAlgorithmClassCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TreeAlgorithmClass


class TreeAlgorithmCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TreeAlgorithm


class LibraryPrepProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.LibraryPrepProtocol


class AssemblyProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AssemblyProtocol


class LocusDetectionProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.LocusDetectionProtocol


class SnpDetectionProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SnpDetectionProtocol


class AlignmentProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AlignmentProtocol


class TaxonomyProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TaxonomyProtocol


class SeqClassificationProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqClassificationProtocol


class PcrProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.PcrProtocol


class AstProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AstProtocol


class SeqDistanceProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqDistanceProtocol


class SeqCategorySetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqCategorySet


class SeqCategoryCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqCategory


class SubtypingSchemeCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SubtypingScheme


class TaxonCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Taxon


class TaxonSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TaxonSet


class TaxonSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TaxonSetMember


class LocusCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Locus


class LocusSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.LocusSet


class LocusSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.LocusSetMember


class RefAlleleCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefAllele


class TaxonLocusLinkCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TaxonLocusLink


class RefSeqCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefSeq


class AlleleCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Allele


class AlleleAlignmentCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AlleleAlignment


class SampleCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Sample


class ReadSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ReadSet


class RawSeqCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RawSeq


class SeqCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Seq


class SeqAlignmentCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqAlignment


class AlleleProfileCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AlleleProfile


class RefSnpCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefSnp


class RefSnpCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefSnp


class RefSnpSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefSnpSet


class RefSnpSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefSnpSetMember


class SeqTaxonomyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqTaxonomy


class PcrMeasurementCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.PcrMeasurement


class AstMeasurementCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AstMeasurement


class AstPredictionCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AstPrediction


class SeqDistanceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqDistance


class SeqClassificationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqClassification


class SnpProfileCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SnpProfile


class KmerProfileCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.KmerProfile


class KmerDetectionProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.KmerDetectionProtocol



# pylint: disable=useless-import-alias
from gen_epix.transform.transformers.conditional import (
    ConditionalTransformer as ConditionalTransformer,
)
from gen_epix.transform.transformers.field import FieldTransformer as FieldTransformer
from gen_epix.transform.transformers.interval import (
    IntervalTransformer as IntervalTransformer,
)
from gen_epix.transform.transformers.iso_time import (
    IsoTimeTransformer as IsoTimeTransformer,
)
from gen_epix.transform.transformers.multi_field import (
    MultiFieldTransformer as MultiFieldTransformer,
)
from gen_epix.transform.transformers.object import (
    ObjectTransformer as ObjectTransformer,
)
from gen_epix.transform.transformers.tuple_map import (
    TupleMapTransformer as TupleMapTransformer,
)
from gen_epix.transform.transformers.validation import (
    ValidationTransformer as ValidationTransformer,
)



import hashlib
import json
import sys
from collections.abc import Callable, Hashable, Iterable
from io import BytesIO, TextIOWrapper
from typing import Any
from uuid import UUID

import numpy as np
import scipy
from Bio import SeqIO
from Bio.Phylo.BaseTree import Clade
from Bio.Phylo.TreeConstruction import DistanceMatrix, DistanceTreeConstructor
from scipy.cluster.hierarchy import ClusterNode

from gen_epix.fastapp import BaseUnitOfWork, CrudOperation, CrudOperationSet
from gen_epix.filter import (
    CompositeFilter,
    EqualsUuidFilter,
    Filter,
    LogicalOperator,
    UuidSetFilter,
)
from gen_epix.seqdb.domain import command, enum, exc, model
from gen_epix.seqdb.domain.repository import BaseSeqRepository
from gen_epix.seqdb.domain.service import BaseSeqService


class SeqService(BaseSeqService):

    def crud(  # type: ignore
        self, cmd: command.CrudCommand
    ) -> list[model.Model] | model.Model | list[UUID] | UUID:
        """
        Override the base crud method to side effects and cascade delete
        where necessary
        """

        # TODO: remove this function once all commands are implemented
        def _get_not_implemented_message(cmd: command.CrudCommand) -> str:
            return (
                f"Command {cmd.__class__.__name__} operation {cmd.operation.value} not implemented for user with role(s) "
                + ", ".join([str(x) for x in cmd.user.roles])
            )

        def _compose_id_filter(*key_and_ids: tuple[str, set[UUID]]) -> Filter:
            return CompositeFilter(
                filters=[
                    UuidSetFilter(key=key, members=frozenset(ids))
                    for key, ids in key_and_ids
                ],
                operator=LogicalOperator.AND,
            )

        # Initialise some
        is_create = cmd.operation in CrudOperationSet.CREATE.value
        is_read = cmd.operation in CrudOperationSet.READ_OR_EXISTS.value
        is_update = cmd.operation in CrudOperationSet.UPDATE.value
        is_delete = cmd.operation in CrudOperationSet.DELETE.value
        access_filter = None

        # Start unit of work and execute all within this scope
        with self.repository.uow() as uow:

            if isinstance(cmd, command.AlleleProfileCrudCommand):
                if is_create:
                    # Calculate all distances for these allele profiles between themselves and with all stored allele profiles
                    allele_profiles: list[model.AlleleProfile] = cmd.get_objs()
                    self._calculate_allele_profile_distances(uow, allele_profiles)

                elif is_read:
                    # Nothing to do extra
                    pass
                elif is_update:
                    # May only change the representation format, not the profile itself
                    raise NotImplementedError(_get_not_implemented_message(cmd))
                elif is_delete:
                    # Delete all distances for these allele profiles as well
                    raise NotImplementedError(_get_not_implemented_message(cmd))
                else:
                    raise NotImplementedError(_get_not_implemented_message(cmd))

        return super().crud(cmd)

    def _calculate_allele_profile_distances(
        self, uow: BaseUnitOfWork, allele_profiles: list[model.AlleleProfile]
    ) -> list[model.SeqDistance]:
        """
        Calculate all distances for these allele profiles between themselves and with
        all stored allele profiles, for all distance protocols that are applicable to
        the locus set of the allele profiles.
        """
        locus_set_ids = {x.locus_set_id for x in allele_profiles}
        cmd = command.SeqDistanceProtocolCrudCommand(
            user=None,
            operation=CrudOperation.READ_ALL,
            query_filter=UuidSetFilter(key="locus_set_id", members=locus_set_ids),
        )
        seq_distance_protocols = self.crud_repository(uow, cmd)
        seq_distances = self.calculate_pairwise_allele_profile_distances(
            seq_distance_protocols, allele_profiles
        )
        # TODO: calculate distances with all stored allele profiles
        # TODO: store/update distances
        # raise NotImplementedError()
        return seq_distances

    def retrieve_phylogenetic_tree(
        self, cmd: command.RetrievePhylogeneticTreeCommand
    ) -> model.PhylogeneticTree | None:
        # profiler = pyinstrument.Profiler(async_mode="enabled")
        # profiler.start()

        user_id = cmd.user.id if cmd.user else None
        seq_ids = cmd.seq_ids
        tree_algorithm = cmd.tree_algorithm
        seq_distance_protocol_id = cmd.seq_distance_protocol_id
        if len(set(seq_ids)) != len(seq_ids):
            raise exc.InvalidArgumentsError("seq_ids must be unique")
        leaf_names = cmd.leaf_names if cmd.leaf_names else [str(x) for x in seq_ids]

        # Retrieve genetic distance protocol
        with self.repository.uow() as uow:
            seq_distance_protocol: model.SeqDistanceProtocol = self.repository.crud(  # type: ignore[assignment]
                uow,
                user_id,
                model.SeqDistanceProtocol,
                None,
                seq_distance_protocol_id,
                CrudOperation.READ_ONE,
            )

        # Special case: 0 or 1 sequences
        if len(seq_ids) < 2:
            return model.PhylogeneticTree(
                id=self.generate_id(),
                tree_algorithm=tree_algorithm,
                seq_distance_protocol_id=seq_distance_protocol_id,
                seq_ids=seq_ids,
                leaf_names=leaf_names,
                newick_repr=f"({leaf_names[0]});" if seq_ids else "();",
            )

        # Retrieve distance matrix
        if tree_algorithm in enum.TreeAlgorithmSet.DISTANCE_BASED.value:
            with self.repository.uow() as uow:
                seq_distances_: list[model.SeqDistance] = self.repository.crud(  # type: ignore[assignment]
                    uow,
                    user_id,
                    model.SeqDistance,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                    filter=CompositeFilter(
                        filters=[
                            UuidSetFilter(key="seq_id", members=frozenset(seq_ids)),
                            EqualsUuidFilter(
                                key="seq_distance_protocol_id",
                                value=seq_distance_protocol_id,
                            ),
                        ],
                        operator=LogicalOperator.AND,
                    ),
                )
                seq_distances = {x.seq_id: x for x in seq_distances_}
            max_stored_distance = seq_distance_protocol.max_stored_distance
            # Calculate condensed distance matrix
            tree_seq_distances_ = [
                seq_distances[x] for x in seq_ids if x in seq_distances
            ]
            tree_leaf_names = [
                x for x, y in zip(leaf_names, seq_ids) if y in seq_distances
            ]
            tree_seq_ids = [x.seq_id for x in tree_seq_distances_]
            tree_seq_ids_index_map = {x: i for i, x in enumerate(tree_seq_ids)}
            str_seq_profile_id_index_map = {
                str(
                    x.allele_profile_id
                    if x.allele_profile_id
                    else (x.snp_profile_id if x.snp_profile_id else x.kmer_profile_id)
                ): tree_seq_ids_index_map[x.seq_id]
                for x in seq_distances.values()
            }
            n_seqs_with_distances = len(tree_seq_ids)
            condensed_distance_matrix = max_stored_distance * np.ones(
                (int(n_seqs_with_distances * (n_seqs_with_distances - 1) / 2),),
                dtype=float,
            )

            def _get_condensed_distance_matrix_index(i: int, j: int, n: int) -> int:
                if i < j:
                    i, j = j, i
                return n * j - j * (j + 1) // 2 + i - 1 - j

            for i, seq_distance in enumerate(tree_seq_distances_):
                if (
                    seq_distance.distance_format
                    != enum.SeqDistanceFormat.SEQ_ID_DISTANCE_DICT
                ):
                    distances = json.loads(seq_distance.distances)
                    for str_seq_profile_id, distance in distances.items():
                        if str_seq_profile_id not in str_seq_profile_id_index_map:
                            # Distance to a sequence not in the list of seq_ids
                            continue
                        j = str_seq_profile_id_index_map[str_seq_profile_id]
                        if distance > max_stored_distance:
                            # Go only up to max_stored_distance in distance matrix,
                            # even if this actual stored distance is larger, e.g.
                            # because the max_stored_distance was higher in the past
                            # TODO: this should be parameterised, so that such higher
                            # distances would nonetheless be used
                            distance = max_stored_distance
                        k = _get_condensed_distance_matrix_index(
                            i, j, n_seqs_with_distances
                        )
                        condensed_distance_matrix[k] = distance
                else:
                    raise exc.InvalidArgumentsError(
                        "Only distance format SEQ_ID_DISTANCE_DICT is supported"
                    )
            # Handle sequences with no stored distances
            if len(tree_seq_ids) < 2:
                return model.PhylogeneticTree(
                    id=self.generate_id(),
                    tree_algorithm=tree_algorithm,
                    seq_distance_protocol_id=seq_distance_protocol_id,
                    seq_ids=seq_ids,
                    leaf_names=leaf_names,
                    newick_repr=f"({tree_leaf_names[0]});" if tree_seq_ids else "();",
                )
            # Calculate tree
            # Increase recursion limit to allow for larger trees
            sys_recursion_limit = sys.getrecursionlimit()
            sys.setrecursionlimit(sys_recursion_limit + len(tree_seq_ids) + 1)
            scipy_tree_algorithm_code_map = {
                enum.TreeAlgorithm.SLINK: "single",
                enum.TreeAlgorithm.UPGMA: "average",
            }
            try:
                if tree_algorithm in scipy_tree_algorithm_code_map:
                    linkage_result = scipy.cluster.hierarchy.linkage(
                        condensed_distance_matrix,
                        scipy_tree_algorithm_code_map[tree_algorithm],
                    )
                    tree = scipy.cluster.hierarchy.to_tree(linkage_result, False)
                    newick_repr = SeqService._get_newick_repr_recursion(
                        tree, tree.dist, tree_leaf_names
                    )
                elif tree_algorithm == enum.TreeAlgorithm.NJ:
                    # TODO: convert condensed distance matrix directly to lower triangle
                    distance_matrix = scipy.spatial.distance.squareform(
                        condensed_distance_matrix
                    )
                    lower_triangle = []
                    for i, x in enumerate(distance_matrix):
                        lower_triangle.append(list(x[: i + 1]))
                    names = [str(x) for x in tree_leaf_names]
                    distance_tree_constructor = DistanceTreeConstructor()
                    bio_distance_matrix = DistanceMatrix(names, lower_triangle)
                    tree = distance_tree_constructor.nj(bio_distance_matrix)
                    # Neighbour joining can produce negative branch lengths
                    # https://en.wikipedia.org/wiki/Neighbor_joining
                    # https://www.researchgate.net/post/How-to-correct-negative-branches-from-neighbor-joining-method
                    # These are corrected here by adding the single negative minimum branch length to all branches
                    SeqService._correct_nj_tree_negative_branch_lengths_recursion(
                        tree.clade
                    )
                    newick_repr = tree.format("newick")
                else:
                    raise exc.InvalidArgumentsError(
                        f"{tree_algorithm.value} tree algorithm not yet implemented"
                    )
            finally:
                # Always set recursion limit back to allow for larger trees
                sys.setrecursionlimit(sys_recursion_limit)
        else:
            raise exc.InvalidArgumentsError(
                f"{tree_algorithm.value} tree algorithm not yet implemented"
            )
        phylogenetic_tree = model.PhylogeneticTree(
            id=self.generate_id(),
            tree_algorithm=tree_algorithm,
            seq_distance_protocol_id=seq_distance_protocol_id,
            seq_ids=seq_ids,
            leaf_names=leaf_names,
            newick_repr=newick_repr,
        )
        # profiler.stop()
        # profiler.write_html(
        #     "./test/output/profile_retrieve_phylogenetic_tree.html"
        # )
        return phylogenetic_tree

    def retrieve_allele_profile(
        self,
        cmd: command.RetrieveCompleteAlleleProfileCommand,
    ) -> model.CompleteAlleleProfile | list[model.CompleteAlleleProfile]:
        raise NotImplementedError()

    def retrieve_snp_profile(
        self, cmd: command.RetrieveCompleteSnpProfileCommand
    ) -> model.CompleteSnpProfile | list[model.CompleteSnpProfile]:
        raise NotImplementedError()

    def retrieve_contig(
        self, cmd: command.RetrieveCompleteContigCommand
    ) -> model.CompleteContig | list[model.CompleteContig]:
        raise NotImplementedError()

    def retrieve_multiple_alignment(
        self, cmd: command.RetrieveMultipleAlignmentCommand
    ) -> model.MultipleAlignment | list[model.MultipleAlignment]:
        raise NotImplementedError()

    def retrieve_sample(
        self, cmd: command.RetrieveCompleteSampleCommand
    ) -> model.CompleteSample | list[model.CompleteSample]:
        raise NotImplementedError()

    def retrieve_seq(
        self, cmd: command.RetrieveCompleteSeqCommand
    ) -> model.CompleteSeq | list[model.CompleteSeq]:
        raise NotImplementedError()

    def retrieve_seq_fasta(self, cmd: command.RetrieveSeqFastaCommand) -> Iterable[str]:
        """
        Retrieve the raw sequences for the given sequence IDs in FASTA format
        as an iterable that yields one sequence at a time.
        """
        wrap = cmd.wrap or 0
        self.repository: BaseSeqRepository
        with self.repository.uow() as uow:
            for seq_id, raw_seq in self.repository.retrieve_seq_fasta(uow, cmd.seq_ids):
                header = f">{seq_id}\n"
                if not wrap:
                    yield f"{header}{raw_seq}\n"
                seq_length = len(raw_seq)
                n_chunks = (seq_length // wrap) + (seq_length % wrap > 0)
                yield header + "\n".join(
                    raw_seq[i * wrap : min((i + 1) * wrap, seq_length)]
                    for i in range(n_chunks)
                )

    @staticmethod
    def calculate_pairwise_allele_profile_distances(
        seq_distance_protocols: Iterable[model.SeqDistanceProtocol],
        allele_profiles: Iterable[model.AlleleProfile],
    ) -> list[model.SeqDistance]:
        """
        Calculate all distances for a set of allele profiles between themselves for all
        the given distance protocols.
        """
        seq_distances: list[model.SeqDistance] = []
        # Go over each distance protocol
        for seq_distance_protocol in seq_distance_protocols:
            assert seq_distance_protocol.id is not None
            locus_set_id = seq_distance_protocol.locus_set_id
            if locus_set_id is None:
                raise exc.InvalidArgumentsError(
                    "SeqDistanceProtocol must have a locus_set_id"
                )
            # Get distance calculation function
            if (
                seq_distance_protocol.seq_distance_protocol_type
                == enum.SeqDistanceProtocolType.ALLELE_HAMMING
            ):
                calculate_distance = SeqService.calculate_hamming_distance
            else:
                raise NotImplementedError()
            # Select only allele profiles for this locus set that are of usable quality
            curr_allele_profiles: list[model.AlleleProfile] = [
                x
                for x in allele_profiles
                if x.locus_set_id == locus_set_id
                and x.quality
                and x.quality.is_usable()
            ]
            # Convert allele_profile from json to object
            allele_profile_allele_ids = [
                json.loads(x.allele_profile) for x in curr_allele_profiles
            ]
            allele_profile_str_seq_ids = [str(x.seq_id) for x in curr_allele_profiles]
            # Go over each unique pair of allele profiles
            curr_seq_distances: dict[int, dict[str, float]] = {
                i: dict() for i in range(len(curr_allele_profiles))
            }
            for i, allele_profile1 in enumerate(curr_allele_profiles):
                # First allele profile
                allele_profile_format1 = allele_profile1.allele_profile_format
                allele_ids1 = allele_profile_allele_ids[i]
                seq_id1 = allele_profile_str_seq_ids[i]
                for j in range(i + 1, len(curr_allele_profiles)):
                    # Second allele profile
                    allele_profile2 = curr_allele_profiles[j]
                    allele_profile_format2 = allele_profile2.allele_profile_format
                    allele_ids2 = allele_profile_allele_ids[j]
                    seq_id2 = allele_profile_str_seq_ids[j]
                    # Calculate distance depending on format of each allele profile
                    distance = SeqService.calculate_allele_profile_distance(
                        calculate_distance,
                        allele_profile_format1,
                        allele_ids1,
                        allele_profile_format2,
                        allele_ids2,
                    )
                    # Keep only distances up to the maximum
                    if distance > seq_distance_protocol.max_stored_distance:
                        continue
                    # Add to seq_distances
                    curr_seq_distances[i][seq_id2] = distance
                    curr_seq_distances[j][seq_id1] = distance

            # Create SeqDistance objects from distances
            for i, allele_profile in enumerate(curr_allele_profiles):
                # Calculate SeqDistance.id as 128 bit hash of seq_id, so that it is always the same
                seq_distance_id = UUID(
                    bytes=hashlib.md5(allele_profile_str_seq_ids[i].encode()).digest()
                )
                # Create seq_distance and add to dict_db
                seq_distance = model.SeqDistance(
                    id=seq_distance_id,
                    seq_id=allele_profile.seq_id,
                    seq_distance_protocol_id=seq_distance_protocol.id,
                    allele_profile_id=allele_profile.id,
                    distance_format=enum.SeqDistanceFormat.SEQ_ID_DISTANCE_DICT,
                    distances=json.dumps(curr_seq_distances[i]),
                )
                seq_distances.append(seq_distance)

        return seq_distances

    @staticmethod
    def calculate_allele_profile_distance(
        calculate_distance: Callable[[list[Hashable], list[Hashable]], float],
        allele_profile_format1: enum.AlleleProfileFormat,
        allele_ids1: list[Hashable],
        allele_profile_format2: enum.AlleleProfileFormat,
        allele_ids2: list[Hashable],
    ) -> float:
        """
        Calculate the distance between two allele profiles
        """
        if allele_profile_format1 == enum.AlleleProfileFormat.SORTED_ALLELE_IDS:
            if allele_profile_format2 == enum.AlleleProfileFormat.SORTED_ALLELE_IDS:
                distance = calculate_distance(allele_ids1, allele_ids2)
            else:
                raise NotImplementedError()
        else:
            raise NotImplementedError()
        return distance

    @staticmethod
    def calculate_hamming_distance(ids1: list[Hashable], ids2: list[Hashable]) -> float:
        """
        Calculate Hamming distance between allele or snp profiles: per locus, add 1
        to the distance if the alleles are different. In case one of the two loci are
        missing, the distance is not increased and neither is it if both are missing
        """
        return float(
            sum(
                1
                for x, y in zip(ids1, ids2)
                if x != y and x is not None and y is not None
            )
        )

    @staticmethod
    def _correct_nj_tree_negative_branch_lengths_recursion(clade: Clade) -> None:
        """
        Recursively update negative branch lengths by adding the negative branch
        length to all siblings. Only one sibling may have a negative branch length.
        """

        # TODO: check if this is correct. Non-terminal branches may have their length
        # updated (extended). As a result their distance to other clades also
        # increases, even if the distance to the sibling that had the negative branch
        # length remains identical.
        if clade.is_terminal():
            return
        min_branch_length = 0
        for subclade in clade.clades:
            if subclade.branch_length < 0:
                if min_branch_length < 0:
                    raise ValueError("More than one negative branch length in a clade")
                min_branch_length = subclade.branch_length
        if min_branch_length < 0:
            for subclade in clade.clades:
                subclade.branch_length -= min_branch_length
        for subclade in clade.clades:
            SeqService._correct_nj_tree_negative_branch_lengths_recursion(subclade)

    @staticmethod
    def _get_newick_repr_recursion(
        node: ClusterNode, parent_dist: float, leaf_names: list[str], newick: str = ""
    ) -> str:
        """
        Convert sciply.cluster.hierarchy.to_tree()-output to Newick format.

        :param node: output of sciply.cluster.hierarchy.to_tree()
        :param parent_dist: output of sciply.cluster.hierarchy.to_tree().dist
        :param leaf_names: list of leaf names
        :param newick: leave empty, this variable is used in recursion.
        :returns: tree in Newick format
        """
        if node.is_leaf():
            return f"{leaf_names[node.id]}:{parent_dist - node.dist:.2f}{newick}"
        if newick:
            newick = f"):{parent_dist - node.dist:.2f}{newick}"
        else:
            newick = ");"
        newick = SeqService._get_newick_repr_recursion(
            node.get_left(),
            node.dist,
            leaf_names,
            newick=newick,
        )
        newick = SeqService._get_newick_repr_recursion(
            node.get_right(),
            node.dist,
            leaf_names,
            newick=f",{newick}",
        )
        newick = f"({newick}"
        return newick




import abc
from collections.abc import Callable
from enum import Enum
from typing import Any

from dynaconf import Dynaconf

from gen_epix.fastapp import App, BaseService
from gen_epix.fastapp.domain.entity import Entity
from gen_epix.fastapp.repositories.dict.repository import DictRepository
from gen_epix.fastapp.repositories.sa.repository import SARepository
from gen_epix.fastapp.repository import BaseRepository


class BaseAppComposer(abc.ABC):
    def __init__(self) -> None:
        self._cfg: Dynaconf
        self._app: App
        self._services: dict[Enum, BaseService]
        self._repositories: dict[Enum, BaseRepository]
        self._registered_user_dependency: Callable
        self._new_user_dependency: Callable
        self._idp_user_dependency: Callable
        raise NotImplementedError()

    @property
    def cfg(self) -> Dynaconf:
        return self._cfg

    @property
    def app(self) -> App:
        return self._app

    @property
    def services(self) -> dict[Enum, BaseService]:
        return self._services

    @property
    def repositories(self) -> dict[Enum, BaseRepository]:
        return self._repositories

    @property
    def registered_user_dependency(self) -> Callable:
        return self._registered_user_dependency

    @property
    def new_user_dependency(self) -> Callable:
        return self._new_user_dependency

    @property
    def idp_user_dependency(self) -> Callable:
        return self._idp_user_dependency

    @classmethod
    def create_repository(
        cls,
        service_type: Enum,
        timestamp_factory: Callable,
        entities: list[Entity],
        repository_type: Enum,
        repository_cfg: dict[str, Any],
        repository_class: type[BaseRepository],
        **kwargs: Any,
    ) -> BaseRepository:
        repository: BaseRepository
        if repository_type.value == "DICT":
            repository = DictRepository.create_repository_from_pkl(
                repository_class,
                entities,
                repository_cfg["props"]["file"],
                timestamp_factory=timestamp_factory,
                **kwargs,
            )
        elif repository_type.value == "SA_SQLITE":
            assert issubclass(repository_class, SARepository)
            repository = repository_class.create_sa_repository(
                entities,
                "sqlite:///" + repository_cfg["props"]["file"],
                name=service_type.value,
                timestamp_factory=timestamp_factory,
                **kwargs,
            )
        elif repository_type.value == "SA_SQL":
            assert issubclass(repository_class, SARepository)
            repository = repository_class.create_sa_repository(
                entities,
                repository_cfg["props"]["connection_string"],
                name=service_type.value,
                timestamp_factory=timestamp_factory,
                **kwargs,
            )
        else:
            raise NotImplementedError()
        return repository



from gen_epix.casedb.domain import command
from gen_epix.casedb.domain.enum import ServiceType
from gen_epix.casedb.domain.repository import BaseOntologyRepository
from gen_epix.fastapp import BaseService


class BaseOntologyService(BaseService):
    SERVICE_TYPE = ServiceType.ONTOLOGY

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseOntologyRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        f = self.app.register_handler
        self.register_default_crud_handlers()
        for command_class in self.app.domain.get_commands_for_service_type(
            ServiceType.ONTOLOGY, base_class=command.UpdateAssociationCommand
        ):
            f(command_class, self.update_association)



from decimal import Decimal
from typing import Literal

from pydantic import Field

from gen_epix.filter.enum import FilterType
from gen_epix.filter.hashable_set import HashableSetFilter


class NumberSetFilter(HashableSetFilter):
    members: frozenset[int | float | Decimal] = Field(
        default=None, description="The numbers to match.", frozen=True
    )


class TypedNumberSetFilter(NumberSetFilter):
    type: Literal[FilterType.NUMBER_SET.value]



# pylint: disable=useless-import-alias
from gen_epix.casedb.api.case import (
    CreateCaseSetRequestBody as CreateCaseSetRequestBody,
)
from gen_epix.casedb.api.case import CreateCasesRequestBody as CreateCasesRequestBody
from gen_epix.casedb.api.case import (
    RetrieveAlleleProfileRequestBody as RetrieveAlleleProfileRequestBody,
)
from gen_epix.casedb.api.case import (
    RetrieveGeneticSequenceRequestBody as RetrieveGeneticSequenceRequestBody,
)
from gen_epix.casedb.api.case import (
    RetrieveOrganizationContactRequestBody as RetrieveOrganizationContactRequestBody,
)
from gen_epix.casedb.api.case import (
    RetrievePhylogeneticTreeRequestBody as RetrievePhylogeneticTreeRequestBody,
)
from gen_epix.casedb.api.case import (
    UpdateCaseTypeColSetCaseTypeColsRequestBody as UpdateCaseTypeColSetCaseTypeColsRequestBody,
)
from gen_epix.casedb.api.case import (
    UpdateCaseTypeSetCaseTypesRequestBody as UpdateCaseTypeSetCaseTypesRequestBody,
)
from gen_epix.casedb.api.ontology import (
    UpdateDiseaseEtiologicalAgentRequestBody as UpdateDiseaseEtiologicalAgentRequestBody,
)
from gen_epix.commondb.api import (
    UpdateUserOwnOrganizationRequestBody as UpdateUserOwnOrganizationRequestBody,
)
from gen_epix.commondb.api.organization import (
    UpdateUserRequestBody as UpdateUserRequestBody,
)
from gen_epix.commondb.api.organization import (
    UserInvitationRequestBody as UserInvitationRequestBody,
)



from gen_epix.commondb.domain.service.organization import BaseOrganizationService
from gen_epix.commondb.domain.service.rbac import BaseRbacService
from gen_epix.fastapp.user_manager import BaseUserManager as ServiceUserManager


class BaseUserManager(ServiceUserManager):
    DEFAULT_KEY_CLAIM = "__key__"

    DEFAULT_NAME_CLAIMS: list[str | list[str]] = [
        "name",
        ["first_name", "last_name"],
        "preferred_username",
        "preferredUsername",
        "username",
    ]

    def __init__(
        self,
        organization_service: BaseOrganizationService,
        rbac_service: BaseRbacService,
        root_cfg: dict[str, dict[str, str]],
        automatic_new_user_cfg: dict[str, dict[str, str]] | None = None,
        key_claim: str | None = None,
        name_claims: list[str | list[str]] | None = None,
    ):
        # Assign input properties
        self._organization_service = organization_service
        self._rbac_service = rbac_service
        self._key_claim = key_claim or self.DEFAULT_KEY_CLAIM
        self._name_claims = name_claims or self.DEFAULT_NAME_CLAIMS
        self._root_cfg = root_cfg
        self._automatic_new_user_cfg = automatic_new_user_cfg



from typing import Any

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, enum, model
from gen_epix.commondb.domain.policy import BaseReadOrganizationResultsOnlyPolicy
from gen_epix.commondb.domain.service import BaseAbacService
from gen_epix.fastapp import CrudOperation, CrudOperationSet, exc


class ReadOrganizationResultsOnlyPolicy(BaseReadOrganizationResultsOnlyPolicy):

    def __init__(self, abac_service: BaseAbacService, **kwargs: Any):
        super().__init__(abac_service, **kwargs)

        app_impl: AppImplDetails = abac_service.app.impl
        self.user_crud_command_class: type[command.UserCrudCommand] = (
            app_impl.get_mapped_class(command.UserCrudCommand)
        )
        self.role_map = app_impl.role_map
        self.role_set_map = app_impl.role_set_map

        self.has_organization_id_attr_command_classes: set[type[command.Command]] = {
            command.UserCrudCommand,
            command.OrganizationAdminPolicyCrudCommand,
            command.UserInvitationCrudCommand,
        }
        self.has_user_id_attr_command_classes: set[type[command.Command]] = set()

    def filter(self, cmd: command.Command, retval: Any) -> Any:  # type: ignore[override]
        if not cmd.user or not cmd.user.id:
            raise exc.ServiceException("Command has no user")
        # TODO: replace filter for AFTER with injecting a filter DURING for efficiency
        if isinstance(cmd, command.RetrieveInviteUserConstraintsCommand):
            # Already handled DURING
            return retval
        if not isinstance(cmd, command.CrudCommand):
            raise NotImplementedError
        if cmd.operation not in CrudOperationSet.READ_OR_EXISTS.value:
            # Policy only applies to read or exists operations
            return retval

        # Roles exempt from this policy
        is_exempt = (
            len(
                cmd.user.roles.intersection(
                    self.role_set_map[enum.RoleSet.GE_APP_ADMIN]
                )
            )
            > 0
        )
        if is_exempt:
            return retval

        # Get organizations to filter on: user's own organization plus any
        # organizations they are admin for
        organization_ids = self.abac_service.retrieve_organizations_under_admin(
            command.RetrieveOrganizationsUnderAdminCommand(user=cmd.user)
        )
        if organization_ids:
            organization_ids.add(cmd.user.organization_id)
        else:
            organization_ids = {cmd.user.organization_id}
        # Filter results based on organizations
        is_read_all = cmd.operation == CrudOperation.READ_ALL
        is_read_one = cmd.operation == CrudOperation.READ_ONE
        msg1 = "User is not an admin for the organization and/or does not belong to it"
        msg2 = "User is not an admin for some of the organizations and/or does not belong to them"
        for command_class in self.has_organization_id_attr_command_classes:
            if not isinstance(cmd, command_class):
                continue
            if is_read_all:
                return [x for x in retval if x.organization_id in organization_ids]
            if is_read_one and retval.organization_id not in organization_ids:
                raise exc.UnauthorizedAuthError(msg1)
            if not is_read_one and any(
                x.organization_id not in organization_ids for x in retval
            ):
                raise exc.UnauthorizedAuthError(msg2)
        for command_class in self.has_user_id_attr_command_classes:
            if not isinstance(cmd, command_class):
                continue
            objs: list = cmd.get_objs() if not is_read_all else []  # type: ignore[attr-defined]
            users: list[model.User] = self.abac_service.app.handle(
                self.user_crud_command_class(
                    user=cmd.user,
                    objs=None,
                    obj_ids=(None if is_read_all else list({x.user_id for x in objs})),
                    operation=(
                        CrudOperation.READ_ALL
                        if is_read_all
                        else CrudOperation.READ_SOME
                    ),
                )
            )
            valid_user_ids = {
                x.id for x in users if x.organization_id in organization_ids
            }
            if is_read_all:
                return [x for x in retval if x.user_id in valid_user_ids]
            else:
                if is_read_one and retval.user_id not in valid_user_ids:
                    raise exc.UnauthorizedAuthError(msg1)
                if not is_read_one and not {x.user_id for x in retval}.issubset(
                    valid_user_ids
                ):
                    raise exc.UnauthorizedAuthError(msg2)
        raise NotImplementedError



import json
import logging
from collections.abc import Callable
from enum import Enum
from typing import Any, NoReturn

from fastapi import APIRouter, FastAPI
from pydantic import BaseModel as PydanticBaseModel

from gen_epix.commondb.api import exc
from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, enum, model
from gen_epix.commondb.domain.model.system import PackageMetadata
from gen_epix.fastapp import App, LogLevel
from gen_epix.fastapp.api import CrudEndpointGenerator

external_logger_fmap = exc.get_logger_fmap(logging.getLogger("commondb.external"))


class HealthStatus(Enum):
    HEALTHY = "HEALTHY"
    UNHEALTHY = "UNHEALTHY"


class HealthReponseBody(PydanticBaseModel):
    status: HealthStatus


class LogItem(PydanticBaseModel):
    level: LogLevel
    command_id: str
    timestamp: str
    duration: float | None = None
    software_version: str
    topic: str
    detail: str | dict | None = None


class LogRequestBody(PydanticBaseModel):
    log_items: list[LogItem]


class LicensesResponseBody(PydanticBaseModel):
    packages: list[PackageMetadata]


def create_system_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    service_type: enum.ServiceType = enum.ServiceType.SYSTEM,
    handle_exception: Callable[[str, Any, Exception], NoReturn] | None = None,
    **kwargs: Any,
) -> None:

    assert handle_exception
    app_impl: AppImplDetails = app.impl
    registered_user_dependency = app_impl.registered_user_dependency
    idp_user_dependency = app_impl.idp_user_dependency

    # Health endpoint
    @router.get(
        "/health",
        operation_id="health",
        name="Health",
    )
    async def health() -> HealthReponseBody:
        """
        Returns the health status of the service. If no response is received
        within the timeout period, the service is considered unhealthy.
        """
        return HealthReponseBody(
            status=HealthStatus.HEALTHY,
        )

    # Licenses endpoint
    @router.post(
        "/retrieve/licenses",
        operation_id="retrieve__licenses",
        name="Licenses",
        description=command.RetrieveLicensesCommand.__doc__,
    )
    async def licenses(
        idp_user: idp_user_dependency,  # type: ignore
    ) -> list[model.PackageMetadata]:
        try:
            cmd = command.RetrieveLicensesCommand(user=None)
            retval: list[model.PackageMetadata] = app.handle(cmd)
        except Exception as exception:
            handle_exception("6ba2c4ca", None, exception)
        return retval

    # Log
    @router.post("/log", operation_id="log")
    async def log(user: registered_user_dependency, request_body: LogRequestBody) -> None:  # type: ignore
        """
        Logs the provided log items.
        """
        try:
            user_id = str(user.id)  # type: ignore[attr-defined]
            for log_item in request_body.log_items:
                if isinstance(log_item.detail, str):
                    log_item.detail = json.loads(log_item.detail)
                content_str = app.create_log_message(
                    log_item.command_id,
                    None,
                    add_debug_info=False,
                    user_id=user_id,
                    **log_item.model_dump(
                        exclude_none=True, exclude={"level", "command_id"}
                    ),
                )
                external_logger_fmap[log_item.level](content_str)
        except Exception as exception:
            handle_exception("09c8e2cd", user, exception)

    # Outage
    @router.get(
        "/retrieve/outages",
        operation_id="retrieve__outages",
        name="Outages",
        description=command.RetrieveOutagesCommand.__doc__,
    )
    async def retrieve__outages(
        idp_user: idp_user_dependency,  # type: ignore
    ) -> list[model.Outage]:
        try:
            cmd = command.RetrieveOutagesCommand(user=None)
            retval: list[model.Outage] = app.handle(cmd)
        except Exception as exception:
            handle_exception("6b47b8b6", None, exception)
        return retval

    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=service_type,
        user_dependency=registered_user_dependency,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



import datetime
from typing import Literal

from pydantic import Field

from gen_epix.filter.enum import FilterType
from gen_epix.filter.range import RangeFilter


class DateRangeFilter(RangeFilter):
    lower_bound: datetime.date | None = Field(
        default=None, description="The lower bound of the range.", frozen=True
    )
    upper_bound: datetime.date | None = Field(
        default=None, description="The upper bound of the range.", frozen=True
    )


class TypedDateRangeFilter(DateRangeFilter):
    type: Literal[FilterType.DATE_RANGE.value]



from typing import ClassVar

from pydantic import Field

from gen_epix.commondb.domain.model.base import Model
from gen_epix.fastapp.domain import Entity


class File(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="files",
        table_name="file",
        persistable=True,
    )
    content: bytes = Field(description="The content of the file.")



import logging
from collections.abc import Callable

from fastapi import FastAPI, Request, Response
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware

from gen_epix.fastapp.app import App


class HandleNoResponseMiddleware(BaseHTTPMiddleware):
    """
    Middleware to handle cases where no response is returned from the endpoint.
    This is a workaround for a known issue in FastAPI where a RuntimeError is raised
    when the request is disconnected before a response is returned.

    Reference: https://stackoverflow.com/questions/71222144/runtimeerror-no-response-returned-in-fastapi-when-refresh-request
    """

    def __init__(
        # Mandatory parameters put as kwargs to comply with the signature of
        # BaseHTTPMiddleware
        self,
        app: FastAPI,
        fast_app: App = None,  # type: ignore[assignment]
        logger: logging.Logger | None = None,
    ):
        super().__init__(app)
        self._fast_app = fast_app
        self._logger = logger or fast_app.logger

    async def dispatch(self, request: Request, call_next: Callable) -> Response:
        try:
            response: Response = await call_next(request)
            return response
        except RuntimeError as exc:
            if str(exc) == "No response returned." and await request.is_disconnected():
                return JSONResponse(None, status_code=204)
            raise



# pylint: disable=useless-import-alias
from gen_epix.fastapp.middleware.handle_auth_exception import (
    HandleAuthExceptionMiddleware as HandleAuthExceptionMiddleware,
)
from gen_epix.fastapp.middleware.handle_no_response import (
    HandleNoResponseMiddleware as HandleNoResponseMiddleware,
)
from gen_epix.fastapp.middleware.limiter import limiter_key_func as limiter_key_func
from gen_epix.fastapp.middleware.update_response_header import (
    UpdateResponseHeaderMiddleware as UpdateResponseHeaderMiddleware,
)



# pylint: disable=useless-import-alias
from gen_epix.commondb.repositories.sa_model import Contact as Contact
from gen_epix.commondb.repositories.sa_model import DataCollection as DataCollection
from gen_epix.commondb.repositories.sa_model import (
    DataCollectionSet as DataCollectionSet,
)
from gen_epix.commondb.repositories.sa_model import (
    DataCollectionSetMember as DataCollectionSetMember,
)
from gen_epix.commondb.repositories.sa_model import IdentifierIssuer as IdentifierIssuer
from gen_epix.commondb.repositories.sa_model import Organization as Organization
from gen_epix.commondb.repositories.sa_model import (
    OrganizationAdminPolicy as OrganizationAdminPolicy,
)
from gen_epix.commondb.repositories.sa_model import OrganizationSet as OrganizationSet
from gen_epix.commondb.repositories.sa_model import (
    OrganizationSetMember as OrganizationSetMember,
)
from gen_epix.commondb.repositories.sa_model import Outage as Outage
from gen_epix.commondb.repositories.sa_model import RowMetadataMixin
from gen_epix.commondb.repositories.sa_model import Site as Site
from gen_epix.commondb.repositories.sa_model import User as User
from gen_epix.commondb.repositories.sa_model import UserInvitation as UserInvitation
from gen_epix.commondb.repositories.sa_model import (
    create_field_metadata,
    set_entity_repository_model_classes,
)
from gen_epix.omopdb.domain import DOMAIN, enum, model
from gen_epix.omopdb.repositories.sa_model.omop import CareSite as CareSite
from gen_epix.omopdb.repositories.sa_model.omop import CdmSource as CdmSource
from gen_epix.omopdb.repositories.sa_model.omop import Cohort as Cohort
from gen_epix.omopdb.repositories.sa_model.omop import (
    CohortDefinition as CohortDefinition,
)
from gen_epix.omopdb.repositories.sa_model.omop import Concept as Concept
from gen_epix.omopdb.repositories.sa_model.omop import (
    ConceptAncestor as ConceptAncestor,
)
from gen_epix.omopdb.repositories.sa_model.omop import ConceptClass as ConceptClass
from gen_epix.omopdb.repositories.sa_model.omop import (
    ConceptRelationship as ConceptRelationship,
)
from gen_epix.omopdb.repositories.sa_model.omop import ConceptSynonym as ConceptSynonym
from gen_epix.omopdb.repositories.sa_model.omop import ConditionEra as ConditionEra
from gen_epix.omopdb.repositories.sa_model.omop import (
    ConditionOccurrence as ConditionOccurrence,
)
from gen_epix.omopdb.repositories.sa_model.omop import Cost as Cost
from gen_epix.omopdb.repositories.sa_model.omop import DeviceExposure as DeviceExposure
from gen_epix.omopdb.repositories.sa_model.omop import Domain as Domain
from gen_epix.omopdb.repositories.sa_model.omop import DoseEra as DoseEra
from gen_epix.omopdb.repositories.sa_model.omop import DrugEra as DrugEra
from gen_epix.omopdb.repositories.sa_model.omop import DrugExposure as DrugExposure
from gen_epix.omopdb.repositories.sa_model.omop import DrugStrength as DrugStrength
from gen_epix.omopdb.repositories.sa_model.omop import (
    FactRelationship as FactRelationship,
)
from gen_epix.omopdb.repositories.sa_model.omop import Location as Location
from gen_epix.omopdb.repositories.sa_model.omop import (
    LocationHistory as LocationHistory,
)
from gen_epix.omopdb.repositories.sa_model.omop import Measurement as Measurement
from gen_epix.omopdb.repositories.sa_model.omop import (
    MeasurementRelation as MeasurementRelation,
)
from gen_epix.omopdb.repositories.sa_model.omop import Metadata as Metadata
from gen_epix.omopdb.repositories.sa_model.omop import Note as Note
from gen_epix.omopdb.repositories.sa_model.omop import NoteNlp as NoteNlp
from gen_epix.omopdb.repositories.sa_model.omop import Observation as Observation
from gen_epix.omopdb.repositories.sa_model.omop import (
    ObservationPeriod as ObservationPeriod,
)
from gen_epix.omopdb.repositories.sa_model.omop import (
    PayerPlanPeriod as PayerPlanPeriod,
)
from gen_epix.omopdb.repositories.sa_model.omop import Person as Person
from gen_epix.omopdb.repositories.sa_model.omop import (
    ProcedureOccurrence as ProcedureOccurrence,
)
from gen_epix.omopdb.repositories.sa_model.omop import Provider as Provider
from gen_epix.omopdb.repositories.sa_model.omop import Relationship as Relationship
from gen_epix.omopdb.repositories.sa_model.omop import (
    SourceToConceptMap as SourceToConceptMap,
)
from gen_epix.omopdb.repositories.sa_model.omop import Specimen as Specimen
from gen_epix.omopdb.repositories.sa_model.omop import SurveyConduct as SurveyConduct
from gen_epix.omopdb.repositories.sa_model.omop import VisitDetail as VisitDetail
from gen_epix.omopdb.repositories.sa_model.omop import (
    VisitOccurrence as VisitOccurrence,
)
from gen_epix.omopdb.repositories.sa_model.omop import Vocabulary as Vocabulary

SA_MODELS_BY_SERVICE_TYPE: dict[enum.ServiceType, dict[type[model.Model], type]] = {
    enum.ServiceType.ABAC: {
        model.OrganizationAdminPolicy: OrganizationAdminPolicy,
    },
    enum.ServiceType.ORGANIZATION: {
        model.Contact: Contact,
        model.DataCollection: DataCollection,
        model.DataCollectionSet: DataCollectionSet,
        model.DataCollectionSetMember: DataCollectionSetMember,
        model.IdentifierIssuer: IdentifierIssuer,
        model.Organization: Organization,
        model.OrganizationSet: OrganizationSet,
        model.OrganizationSetMember: OrganizationSetMember,
        model.Site: Site,
        model.User: User,
        model.UserInvitation: UserInvitation,
    },
    enum.ServiceType.SYSTEM: {
        model.Outage: Outage,
    },
    enum.ServiceType.OMOP: {
        model.CareSite: CareSite,
        model.CdmSource: CdmSource,
        model.Cohort: Cohort,
        model.CohortDefinition: CohortDefinition,
        model.Concept: Concept,
        model.ConceptAncestor: ConceptAncestor,
        model.ConceptClass: ConceptClass,
        model.ConceptRelationship: ConceptRelationship,
        model.ConceptSynonym: ConceptSynonym,
        model.ConditionEra: ConditionEra,
        model.ConditionOccurrence: ConditionOccurrence,
        model.Cost: Cost,
        model.DeviceExposure: DeviceExposure,
        model.Domain: Domain,
        model.DoseEra: DoseEra,
        model.DrugEra: DrugEra,
        model.DrugExposure: DrugExposure,
        model.DrugStrength: DrugStrength,
        model.FactRelationship: FactRelationship,
        model.Location: Location,
        model.LocationHistory: LocationHistory,
        model.Measurement: Measurement,
        model.MeasurementRelation: MeasurementRelation,
        model.Metadata: Metadata,
        model.Note: Note,
        model.NoteNlp: NoteNlp,
        model.Observation: Observation,
        model.ObservationPeriod: ObservationPeriod,
        model.PayerPlanPeriod: PayerPlanPeriod,
        model.Person: Person,
        model.ProcedureOccurrence: ProcedureOccurrence,
        model.Provider: Provider,
        model.Relationship: Relationship,
        model.SourceToConceptMap: SourceToConceptMap,
        model.Specimen: Specimen,
        model.SurveyConduct: SurveyConduct,
        model.VisitDetail: VisitDetail,
        model.VisitOccurrence: VisitOccurrence,
        model.Vocabulary: Vocabulary,
    },
}

FIELD_NAME_MAP: dict[type, dict[str, str]] = {}

set_entity_repository_model_classes(
    DOMAIN,
    SA_MODELS_BY_SERVICE_TYPE,
    RowMetadataMixin,
    field_name_map=FIELD_NAME_MAP,
)

SERVICE_METADATA_FIELDS, DB_METADATA_FIELDS, GENERATE_SERVICE_METADATA = (
    create_field_metadata(DOMAIN)
)



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


import hashlib
from typing import ClassVar, Self
from uuid import UUID

from pydantic import Field, field_serializer, field_validator, model_validator

from gen_epix.commondb.domain.model.base import Model
from gen_epix.fastapp.domain import Entity, create_keys, create_links
from gen_epix.seqdb.domain import enum
from gen_epix.seqdb.domain.model.seq.base import (
    AlignmentMixin,
    CodeMixin,
    ProtocolMixin,
    QualityMixin,
    SeqMixin,
)
from gen_epix.seqdb.domain.model.seq.metadata import (
    AlignmentProtocol,
    AssemblyProtocol,
    AstProtocol,
    KmerDetectionProtocol,
    LibraryPrepProtocol,
    Locus,
    LocusDetectionProtocol,
    LocusSet,
    PcrProtocol,
    RefSeq,
    SeqCategory,
    SeqClassificationProtocol,
    SnpDetectionProtocol,
    Taxon,
    TaxonomyProtocol,
)


class Allele(Model, SeqMixin, QualityMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="alleles",
        table_name="allele",
        persistable=True,
        keys=create_keys({1: ("locus_id", "seq_hash_sha256")}),
        links=create_links({1: ("locus_id", Locus, "locus")}),
    )
    locus_id: UUID = Field(
        description="The unique identifier for the locus. FOREIGN KEY"
    )
    locus: Locus | None = Field(default=None, description="The locus.")


class AlleleAlignment(Model, AlignmentMixin, QualityMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="allele_alignments",
        table_name="allele_alignment",
        persistable=True,
        keys=create_keys({1: ("ref_allele_id", "allele_id", "alignment_protocol_id")}),
        links=create_links(
            {
                1: ("ref_allele_id", Allele, "ref_allele"),
                2: ("allele_id", Allele, "allele"),
                3: (
                    "alignment_protocol_id",
                    AlignmentProtocol,
                    "alignment_protocol",
                ),
            }
        ),
    )
    ref_allele_id: UUID = Field(
        description="The unique identifier for the reference allele. FOREIGN KEY"
    )
    ref_allele: Allele | None = Field(default=None, description="The reference allele.")
    allele_id: UUID = Field(
        description="The unique identifier for the allele. FOREIGN KEY"
    )
    allele: Allele | None = Field(default=None, description="The allele.")
    alignment_protocol_id: UUID = Field(
        description="The unique identifier for the sequence alignment protocol. FOREIGN KEY"
    )
    alignment_protocol: AlignmentProtocol | None = Field(
        default=None, description="The sequence alignment protocol."
    )


class Sample(Model, CodeMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="samples",
        table_name="sample",
        persistable=True,
        keys=create_keys({1: "code"}),
    )
    props: dict[str, str] = Field(
        default_factory=dict, description="The properties of the sample."
    )


class ReadSet(Model, CodeMixin, QualityMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="read_sets",
        table_name="read_set",
        persistable=True,
        keys=create_keys({1: "code"}),
        links=create_links(
            {
                #!FIXME links with None foreign keys are not supported yet
                # 1: (
                #     "fwd_file_id",
                #     File,
                #     None,
                # ),
                # 2: (
                #     "rev_file_id",
                #     File,
                #     None,
                # ),
                1: (
                    "library_prep_protocol_id",
                    LibraryPrepProtocol,
                    "library_prep_protocol",
                ),
            }
        ),
    )
    library_prep_protocol_id: UUID = Field(
        description="The unique identifier for the library preparation protocol. FOREIGN KEY"
    )
    library_prep_protocol: LibraryPrepProtocol | None = Field(
        default=None, description="The sequencing protocol."
    )
    fwd_uri: str | None = Field(
        default=None,
        description="The URI of the forward read set. In case of single-end reads, this is the only read set.",
    )
    rev_uri: str | None = Field(
        default=None, description="The URI of the reverse read set, if any."
    )
    fwd_file_id: UUID | None = Field(
        default=None,
        description="The unique file identifier for the forward read set. In case of single-end reads, this is the only read set. FOREIGN KEY",
    )
    rev_file_id: UUID | None = Field(
        default=None,
        description="The unique file identifier for the reverse read set, if any.",
    )
    fwd_reads_hash_sha256: bytes | None = Field(
        default=None,
        description="The SHA256 hash of the uncompressed FASTQ file representation of the forward read set.",
        min_length=32,
        max_length=32,
    )
    rev_reads_hash_sha256: bytes | None = Field(
        default=None,
        description="The SHA256 hash of the uncompressed FASTQ file representation of the reverse read set.",
        min_length=32,
        max_length=32,
    )
    sequencing_run_code: str | None = Field(
        description="The code of the sequencing run.", max_length=255, default=None
    )

    @model_validator(mode="after")
    def _validate_model(self) -> Self:
        if self.fwd_uri and self.rev_uri and self.fwd_uri == self.rev_uri:
            raise ValueError("fwd_uri must be different from rev_uri")
        if (
            self.fwd_file_id
            and self.rev_file_id
            and self.fwd_file_id == self.rev_file_id
        ):
            raise ValueError("fwd_file_id must be different from rev_file_id")
        if (
            self.fwd_reads_hash_sha256
            and self.rev_reads_hash_sha256
            and self.fwd_reads_hash_sha256 == self.rev_reads_hash_sha256
        ):
            raise ValueError(
                "fwd_reads_hash_sha256 must be different from rev_reads_hash_sha256"
            )
        return self


class RawSeq(Model, SeqMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="raw_seqs",
        table_name="raw_seq",
        persistable=True,
    )


class Seq(Model, CodeMixin, QualityMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="seqs",
        table_name="seq",
        persistable=True,
        keys=create_keys(
            {
                1: "code",
            }
        ),
        links=create_links(
            {
                1: ("sample_id", Sample, "sample"),
                2: ("read_set_id", ReadSet, "read_set"),
                3: ("read_set2_id", ReadSet, "read_set2"),
                4: ("assembly_protocol_id", AssemblyProtocol, "assembly_protocol"),
                5: ("raw_seq_id", RawSeq, "raw_seq"),
            }
        ),
    )
    sample_id: UUID | None = Field(
        default=None,
        description="The unique identifier for the sample, if available. FOREIGN KEY",
    )
    sample: Sample | None = Field(default=None, description="The sample.")
    read_set_id: UUID | None = Field(
        default=None,
        description="The unique identifier for the single read set used to generate the assembly, if available. FOREIGN KEY",
    )
    read_set: ReadSet | None = Field(default=None, description="The read set.")
    read_set2_id: UUID | None = Field(
        default=None,
        description="The unique identifier for a second read set used to generate the assembly, if more than one. FOREIGN KEY",
    )
    read_set2: ReadSet | None = Field(default=None, description="The second read set.")
    assembly_protocol_id: UUID | None = Field(
        default=None,
        description="The unique identifier for the assembly protocol, if available. FOREIGN KEY",
    )
    assembly_protocol: AssemblyProtocol | None = Field(
        default=None, description="The assembly protocol."
    )
    raw_seq_id: UUID | None = Field(
        default=None,
        description="The unique identifier for the raw sequence, if available. FOREIGN KEY",
    )
    raw_seq: RawSeq | None = Field(default=None, description="The raw sequence.")
    file_id: UUID | None = Field(
        default=None, description="The unique file identifier."
    )

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if self.read_set_id is None:
            if self.read_set2_id is not None:
                raise ValueError(
                    "read_set2_id may only be provided if read_set_id is provided"
                )
        elif self.read_set2_id == self.read_set_id:
            raise ValueError("read_set2_id must be different from read_set_id")
        return self


class ContigAlignment(Model, AlignmentMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="contig_alignments",
        persistable=False,
    )
    ref_seq_id: UUID = Field(
        description="The unique identifier for the reference sequence. FOREIGN KEY"
    )


class SeqAlignment(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="seq_alignments",
        table_name="seq_alignment",
        persistable=True,
        keys=create_keys({1: ("seq_id", "alignment_protocol_id")}),
        links=create_links(
            {
                1: ("seq_id", Seq, "seq"),
                2: (
                    "alignment_protocol_id",
                    AlignmentProtocol,
                    "alignment_protocol",
                ),
            }
        ),
    )
    seq_id: UUID = Field(
        description="The unique identifier for the sequence. FOREIGN KEY"
    )
    seq: Seq = Field(description="The sequence.")
    alignment_protocol_id: UUID = Field(
        description="The unique identifier for the sequence alignment protocol. FOREIGN KEY"
    )
    alignment_protocol: AlignmentProtocol | None = Field(
        default=None, description="The sequence alignment protocol."
    )
    contig_alignments: list[ContigAlignment] = Field(
        description="The contig alignments."
    )


class AlleleProfile(Model, QualityMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="allele_profiles",
        table_name="allele_profile",
        persistable=True,
        keys=create_keys(
            {1: ("seq_id", "locus_set_id", "locus_detection_protocol_id")}
        ),
        links=create_links(
            {
                1: ("seq_id", Seq, "seq"),
                2: ("locus_set_id", LocusSet, "locus_set"),
                3: (
                    "locus_detection_protocol_id",
                    LocusDetectionProtocol,
                    "locus_detection_protocol",
                ),
            }
        ),
    )
    seq_id: UUID = Field(
        description="The unique identifier for the sequence. FOREIGN KEY"
    )
    seq: Seq | None = Field(default=None, description="The sequence.")
    locus_set_id: UUID = Field(
        description="The unique identifier for the locus set. FOREIGN KEY"
    )
    locus_set: LocusSet | None = Field(default=None, description="The locus set.")
    locus_detection_protocol_id: UUID = Field(
        description="The unique identifier for the locus detection protocol. FOREIGN KEY"
    )
    locus_detection_protocol: LocusDetectionProtocol | None = Field(
        default=None, description="The locus detection protocol."
    )
    n_loci: int = Field(description="The number of loci detected.")
    allele_profile: str = Field(
        description="The alleles detected in the sequence for the loci in the locus set."
    )
    allele_profile_format: enum.AlleleProfileFormat = Field(
        default=enum.AlleleProfileFormat.SORTED_ALLELE_IDS,
        description="The representation format of the alleles.",
    )
    allele_profile_hash_sha256: bytes = Field(
        description="The SHA256 hash of the sorted list of allele ids as bytes.",
        min_length=32,
        max_length=32,
    )

    @field_validator("allele_profile_hash_sha256", mode="before")
    def _validate_allele_profile_hash_sha256(cls, value: str | bytes) -> bytes:
        if isinstance(value, str):
            value = bytes.fromhex(value)
        return value

    @staticmethod
    def get_allele_profile_hash_sha256(allele_ids: list[UUID | None]) -> bytes:
        sha256 = hashlib.sha256()
        sha256.update(b"".join(sorted([x.bytes for x in allele_ids if x is not None])))
        return sha256.digest()

    @field_serializer("allele_profile_format", mode="plain")
    def _serialize_snp_profile_format(
        self, value: str | enum.AlleleProfileFormat
    ) -> str:
        if isinstance(value, enum.AlleleProfileFormat):
            return value.value
        return value


class SnpProfile(Model, QualityMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="snp_profiles",
        table_name="snp_profile",
        persistable=True,
        keys=create_keys(
            {
                1: (
                    "seq_id",
                    "ref_seq_id",
                    "snp_detection_protocol_id",
                )
            }
        ),
        links=create_links(
            {
                1: ("seq_id", Seq, "seq"),
                2: ("ref_seq_id", RefSeq, "ref_seq"),
                3: (
                    "snp_detection_protocol_id",
                    SnpDetectionProtocol,
                    "snp_detection_protocol",
                ),
            }
        ),
    )
    seq_id: UUID = Field(
        description="The unique identifier for the sequence. FOREIGN KEY"
    )
    seq: Seq | None = Field(default=None, description="The sequence.")
    ref_seq_id: UUID = Field(
        description="The unique identifier for the reference sequence. FOREIGN KEY"
    )
    ref_seq: RefSeq | None = Field(default=None, description="The reference sequence.")
    snp_detection_protocol_id: UUID = Field(
        description="The unique identifier for the SNP detection protocol. FOREIGN KEY"
    )
    snp_detection_protocol: SnpDetectionProtocol | None = Field(
        default=None, description="The SNP detection protocol."
    )
    snp_profile: str = Field(description="The SNPs detected in the sequence.")
    snp_profile_format: enum.SnpProfileFormat = Field(
        default=enum.SnpProfileFormat.REF_ALN_SEQ,
        description="The representation format of the SNPs.",
    )
    snp_profile_hash_sha256: bytes = Field(
        description="The SHA256 hash of the ASCII lower case reference sequence with all SNPs applied.",
        min_length=32,
        max_length=32,
    )

    @field_validator("snp_profile_hash_sha256", mode="before")
    def _validate_snp_profile_hash_sha256(cls, value: str | bytes) -> bytes:
        if isinstance(value, str):
            value = bytes.fromhex(value)
        return value

    @field_serializer("snp_profile_format", mode="plain")
    def _serialize_snp_profile_format(self, value: str | enum.SnpProfileFormat) -> str:
        if isinstance(value, enum.SnpProfileFormat):
            return value.value
        return value


class KmerProfile(Model, QualityMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="kmer_profiles",
        table_name="kmer_profile",
        persistable=True,
        keys=create_keys(
            {
                1: (
                    "seq_id",
                    "kmer_detection_protocol_id",
                )
            }
        ),
        links=create_links(
            {
                1: ("seq_id", Seq, "seq"),
                2: (
                    "kmer_detection_protocol_id",
                    KmerDetectionProtocol,
                    "kmer_detection_protocol",
                ),
            }
        ),
    )
    seq_id: UUID = Field(
        description="The unique identifier for the sequence. FOREIGN KEY"
    )
    seq: Seq | None = Field(default=None, description="The sequence.")
    kmer_detection_protocol_id: UUID = Field(
        description="The unique identifier for the k-mer detection protocol. FOREIGN KEY"
    )
    kmer_detection_protocol: KmerDetectionProtocol | None = Field(
        default=None, description="The k-mer detection protocol."
    )
    kmer_profile: str = Field(
        description="The k-mers detected in the sequence and their frequency."
    )
    kmer_profile_format: enum.KmerProfileFormat = Field(
        default=enum.KmerProfileFormat.KMER_PROFILE_FORMAT1,
        description="The representation format of the k-mers.",
    )
    kmer_profile_hash_sha256: bytes = Field(
        description="The SHA256 hash of the ASCII sorted k-mers followed by their sorted frequencies as double precision floats.",
        min_length=32,
        max_length=32,
    )


class SeqClassification(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="seq_classifications",
        table_name="seq_classification",
        persistable=True,
        keys=create_keys({1: ("seq_id", "seq_classification_protocol_id")}),
        links=create_links(
            {
                1: ("seq_id", Seq, "seq"),
                2: (
                    "seq_classification_protocol_id",
                    SeqClassificationProtocol,
                    "seq_classification_protocol",
                ),
                3: ("primary_category_id", SeqCategory, "primary_category"),
            }
        ),
    )
    seq_id: UUID = Field(
        description="The unique identifier for the sequence. FOREIGN KEY"
    )
    seq: Seq | None = Field(default=None, description="The sequence.")
    seq_classification_protocol_id: UUID = Field(
        description="The ID of the sequence classification protocol. FOREIGN KEY"
    )
    seq_classification_protocol: SeqClassificationProtocol = Field(
        description="The sequence classification protocol."
    )
    primary_category_id: UUID | None = Field(
        description="The ID of the category. FOREIGN KEY"
    )
    primary_category: SeqCategory = Field(description="The primary category.")
    classification: str = Field(description="The classification of the sequence.")
    classification_format: enum.SeqClassificationFormat = Field(
        default=enum.SeqClassificationFormat.SEQ_CLASSIFICATION_FORMAT1,
        description="The representation format of the classification.",
    )
    classification_hash_sha256: bytes = Field(
        description="The SHA256 hash of the sorted list of category ids as bytes.",
        min_length=32,
        max_length=32,
    )


class SeqTaxonomy(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="seq_taxonomies",
        table_name="seq_taxonomy",
        persistable=True,
        keys=create_keys({1: "seq_id", 2: "taxonomy_protocol_id"}),
        links=create_links(
            {
                1: ("seq_id", Seq, "seq"),
                2: ("taxonomy_protocol_id", TaxonomyProtocol, "taxonomy_protocol"),
                3: ("primary_taxon_id", Taxon, "primary_taxon"),
            }
        ),
    )
    seq_id: UUID = Field(
        description="The unique identifier for the sequence. FOREIGN KEY"
    )
    seq: Seq | None = Field(default=None, description="The sequence.")
    taxonomy_protocol_id: UUID = Field(
        description="The unique identifier for the taxonomy protocol. FOREIGN KEY"
    )
    taxonomy_protocol: TaxonomyProtocol | None = Field(
        default=None, description="The taxonomy protocol."
    )
    primary_taxon_id: UUID = Field(
        description="The unique identifier for the primary taxon. FOREIGN KEY"
    )
    primary_taxon: Taxon | None = Field(default=None, description="The primary taxon.")
    taxonomy: str = Field(description="The taxonomy results of the sequence.")
    taxonomy_format: enum.TaxonomyFormat = Field(
        default=enum.TaxonomyFormat.TAXONOMY_FORMAT1,
        description="The representation format of the taxonomy.",
    )
    taxonomy_hash_sha256: bytes = Field(
        description="The SHA256 hash of the sorted list of taxon ids as bytes.",
        min_length=32,
        max_length=32,
    )


class PcrMeasurement(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="pcr_measurements",
        table_name="pcr_measurement",
        persistable=True,
        keys=create_keys({1: ("sample_id", "pcr_protocol_id", "index")}),
        links=create_links(
            {
                1: ("sample_id", Sample, "sample"),
                2: ("pcr_protocol_id", PcrProtocol, "pcr_protocol"),
            }
        ),
    )
    sample_id: UUID = Field(
        description="The unique identifier for the sample. FOREIGN KEY"
    )
    sample: Sample | None = Field(default=None, description="The sample.")
    pcr_protocol_id: UUID = Field(
        description="The unique identifier for the PCR protocol. FOREIGN KEY"
    )
    pcr_protocol: PcrProtocol | None = Field(
        default=None, description="The PCR protocol."
    )
    pcr_result: str = Field(description="The result of the PCR experiment.")
    pcr_result_format: enum.PcrResultFormat = Field(
        default=enum.PcrResultFormat.PCR_RESULT_FORMAT1,
        description="The representation format of the PCR result.",
    )
    index: int = Field(default=1, description="The index of the measurement.")


class AstMeasurement(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ast_measurements",
        table_name="ast_measurement",
        persistable=True,
        keys=create_keys({1: ("sample_id", "ast_protocol_id", "index")}),
        links=create_links(
            {
                1: ("sample_id", Sample, "sample"),
                2: ("ast_protocol_id", AstProtocol, "ast_protocol"),
            }
        ),
    )
    sample_id: UUID = Field(
        description="The unique identifier for the sample. FOREIGN KEY"
    )
    sample: Sample | None = Field(default=None, description="The sample.")
    ast_protocol_id: UUID = Field(
        description="The unique identifier for the AST protocol. FOREIGN KEY"
    )
    ast_protocol: AstProtocol | None = Field(
        default=None, description="The AST protocol."
    )
    ast_result: str = Field(description="The result of the AST experiment.")
    ast_result_format: enum.AstResultFormat = Field(
        default=enum.AstResultFormat.AST_RESULT_FORMAT1,
        description="The representation format of the AST result.",
    )
    index: int = Field(default=1, description="The index of the measurement.")


class AstPrediction(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ast_predictions",
        table_name="ast_prediction",
        persistable=True,
        keys=create_keys({1: ("seq_id", "ast_protocol_id")}),
        links=create_links(
            {
                1: ("seq_id", Seq, "seq"),
                2: ("ast_protocol_id", AstProtocol, "ast_protocol"),
            }
        ),
    )
    seq_id: UUID = Field(
        description="The unique identifier for the sequence. FOREIGN KEY"
    )
    seq: Seq | None = Field(default=None, description="The sequence.")
    ast_protocol_id: UUID = Field(
        description="The unique identifier for the AST protocol. FOREIGN KEY"
    )
    ast_protocol: AstProtocol | None = Field(
        default=None, description="The AST protocol."
    )
    ast_result: str = Field(description="The result of the AST prediction.")
    ast_result_format: enum.AstResultFormat = Field(
        default=enum.AstResultFormat.AST_RESULT_FORMAT1,
        description="The representation format of the AST result.",
    )


class SeqDistanceProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="seq_distance_protocols",
        table_name="seq_distance_protocol",
        persistable=True,
        keys=create_keys(
            {
                1: "code",
                2: ("name", "version"),
            }
        ),
        links=create_links(
            {
                1: ("locus_set_id", LocusSet, "locus_set"),
                2: ("ref_seq_id", RefSeq, "ref_seq"),
            }
        ),
    )
    max_stored_distance: float = Field(description="The maximum distance to be stored")
    min_scale_unit: float = Field(description="The minimum unit to be shown in a scale")
    seq_distance_protocol_type: enum.SeqDistanceProtocolType = Field(
        description="The type of genetic distance protocol.",
    )
    locus_set_id: UUID | None = Field(
        default=None,
        description="The unique identifier for the locus set, if applicable. FOREIGN KEY",
    )
    locus_set: LocusSet | None = Field(default=None, description="The locus set.")
    ref_seq_id: UUID | None = Field(
        default=None,
        description="The unique identifier for the reference sequence, if applicable. FOREIGN KEY",
    )
    ref_seq: RefSeq | None = Field(default=None, description="The reference sequence.")

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if (
            self.seq_distance_protocol_type
            in enum.SeqDistanceProtocolTypeSet.ALLELE_BASED.value
            and self.locus_set_id is None
        ):
            raise ValueError("locus_set_id must be provided for allele based type")
        elif (
            self.seq_distance_protocol_type
            in enum.SeqDistanceProtocolTypeSet.SNP_BASED.value
            and self.ref_seq_id is None
        ):
            raise ValueError("ref_seq_id must be provided for snp based type")
        return self

    @field_serializer("seq_distance_protocol_type", mode="plain")
    def _serialize_seq_format(self, value: str | enum.SeqDistanceProtocolType) -> str:
        if isinstance(value, enum.SeqDistanceProtocolType):
            return value.value
        return value


class SeqDistance(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="seq_distances",
        table_name="seq_distance",
        persistable=True,
        keys=create_keys(
            {
                1: (
                    "seq_id",
                    "seq_distance_protocol_id",
                    "allele_profile_id",
                    "snp_profile_id",
                    "kmer_profile_id",
                )
            }
        ),
        links=create_links(
            {
                1: ("seq_id", Seq, "seq"),
                2: (
                    "seq_distance_protocol_id",
                    SeqDistanceProtocol,
                    "seq_distance_protocol",
                ),
                3: ("allele_profile_id", AlleleProfile, "allele_profile"),
                4: ("snp_profile_id", SnpProfile, "snp_profile"),
                5: ("kmer_profile_id", KmerProfile, "kmer_profile"),
            }
        ),
    )
    seq_id: UUID = Field(
        description="The unique identifier for the sequence. FOREIGN KEY"
    )
    seq: Seq | None = Field(default=None, description="The sequence.")
    seq_distance_protocol_id: UUID = Field(
        description="The unique identifier for the genetic distance protocol. FOREIGN KEY"
    )
    seq_distance_protocol: SeqDistanceProtocol | None = Field(
        default=None, description="The genetic distance protocol."
    )
    allele_profile_id: UUID | None = Field(
        default=None,
        description="The unique identifier for the allele profile, if applicable. FOREIGN KEY",
    )
    allele_profile: AlleleProfile | None = Field(
        default=None, description="The allele profile."
    )
    snp_profile_id: UUID | None = Field(
        default=None,
        description="The unique identifier for the SNP profile, if applicable. FOREIGN KEY",
    )
    snp_profile: SnpProfile | None = Field(default=None, description="The SNP profile.")
    kmer_profile_id: UUID | None = Field(
        default=None,
        description="The unique identifier for the k-mer profile, if applicable. FOREIGN KEY",
    )
    kmer_profile: SnpProfile | None = Field(
        default=None, description="The k-mer profile."
    )
    distance_format: enum.SeqDistanceFormat = Field(
        default=enum.SeqDistanceFormat.SEQ_ID_DISTANCE_DICT,
        description="The representation format of the distances.",
    )
    distances: str = Field(description="The distances to other sequences.")

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        ids = [self.allele_profile_id, self.snp_profile_id, self.kmer_profile_id]
        has_ids = [x is not None for x in ids]
        if not any(has_ids):
            raise ValueError(
                "Either allele_profile_id, snp_profile_id or kmer_profile_id must be provided"
            )
        elif sum(has_ids) > 1:
            raise ValueError(
                "Only one of allele_profile_id, snp_profile_id or kmer_profile_id must be provided"
            )
        objs = [self.allele_profile, self.snp_profile, self.kmer_profile]
        for has_id, obj in zip(has_ids, objs):
            if not has_id and obj is not None:
                raise ValueError(f"{obj.__class__.__name__} must be None")
        return self

    @field_serializer("distance_format", mode="plain")
    def _serialize_distance_format(self, value: str | enum.SeqDistanceFormat) -> str:
        if isinstance(value, enum.SeqDistanceFormat):
            return value.value
        return value



from enum import Enum

from pydantic import BaseModel

from gen_epix.commondb.util import copy_model_field
from gen_epix.fastapp.enum import PermissionType
from gen_epix.fastapp.model import Permission
from gen_epix.seqdb.domain import DOMAIN

CommandName = Enum("CommandName", {x: x for x in DOMAIN.command_names})  # type: ignore[misc] # Dynamic Enum required


class ApiPermission(BaseModel, frozen=True):
    command_name: CommandName = (  # pyright: ignore[reportInvalidTypeForm]
        copy_model_field(Permission, "command_name")
    )
    permission_type: PermissionType = copy_model_field(Permission, "permission_type")



from typing import Literal
from uuid import UUID

from pydantic import Field

from gen_epix.filter.enum import FilterType
from gen_epix.filter.equals import EqualsFilter


class EqualsUuidFilter(EqualsFilter):
    value: UUID = Field(description="The UUID to match.", frozen=True)


class TypedEqualsUuidFilter(EqualsUuidFilter):
    type: Literal[FilterType.EQUALS_UUID.value]



import logging
from collections.abc import Callable
from typing import Any, NoReturn

from dynaconf import Dynaconf
from fastapi import FastAPI, Response
from fastapi.concurrency import asynccontextmanager
from fastapi.responses import RedirectResponse
from slowapi import _rate_limit_exceeded_handler
from slowapi.errors import RateLimitExceeded
from slowapi.middleware import SlowAPIMiddleware

from gen_epix.commondb.api.exc import generate_handle_exception_function
from gen_epix.commondb.api.router import create_routers
from gen_epix.fastapp import App
from gen_epix.fastapp.api.openapi import create_custom_openapi_function
from gen_epix.fastapp.middleware import limiter
from gen_epix.fastapp.middleware.handle_auth_exception import (
    HandleAuthExceptionMiddleware,
)
from gen_epix.fastapp.middleware.update_response_header import (
    UpdateResponseHeaderMiddleware,
)


def create_fast_api(
    app: App,
    create_routers_fn: Callable = create_routers,
    handle_exception: Callable[[str, Any, Exception], NoReturn] | None = None,
    setup_logger: logging.Logger | None = None,
    api_logger: logging.Logger | None = None,
    debug: bool = False,
    **kwargs: Any,
) -> FastAPI:

    cfg: dict | Dynaconf = app.cfg

    # Set up lifespan
    @asynccontextmanager
    async def lifespan(fast_api: FastAPI) -> Any:
        if setup_logger:
            setup_logger.info(
                app.create_log_message(
                    "a49dedfc",
                    {"status": "STARTED_APP"},  # type: ignore[arg-type]
                )
            )
        yield
        if setup_logger:
            setup_logger.info(
                app.create_log_message(
                    "dcabb0ac",
                    {"status": "STOPPING_APP"},  # type: ignore[arg-type]
                )
            )

    # Initialize fast_api
    fast_api = FastAPI(
        separate_input_output_schemas=False,
        swagger_ui_init_oauth={"usePkceWithAuthorizationCodeGrant": True},
        openapi_tags=kwargs.get(
            "openapi_tags",
            [
                {
                    "name": "core",
                    "description": "Core functionality",
                }
            ],
        ),
        lifespan=lifespan,
    )

    # Add middleware
    # Rate limiting
    if not debug:
        fast_api.state.limiter = limiter.limiter
        fast_api.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)  # type: ignore[arg-type]
        # The SlowAPIMiddleware is added to fast_api globally to limit the number of requests
        # The limiter can be applied to specific routes by adding the decorator @limiter.limit
        fast_api.add_middleware(SlowAPIMiddleware)

    # Response header handling
    if not debug:
        exception_headers: list[tuple[set[str], dict[str, str]]] = [
            (
                {"/docs/oauth2-redirect"},
                dict(cfg["api"]["http_header"]["auth"]),
            ),
            (
                {"/docs", "/redoc"},
                dict(cfg["api"]["http_header"]["openapi"]),
            ),
        ]
        fast_api.add_middleware(
            UpdateResponseHeaderMiddleware,
            general_headers=dict(cfg["api"]["http_header"]["general"]),
            exception_headers=exception_headers,
        )
    # Handling of authentication exceptions
    if not debug:
        fast_api.add_middleware(
            HandleAuthExceptionMiddleware,
            fast_app=app,
            logger=api_logger,
        )

    # Add routers
    handle_exception = generate_handle_exception_function(app=app, logger=api_logger)
    routers = create_routers_fn(
        app=app,
        handle_exception=handle_exception,
    )
    for router in routers:
        fast_api.include_router(router, prefix=cfg["api"]["route"]["v1"])

    # Redirect root to default route
    @fast_api.get("/")
    async def redirect() -> Response:
        response = RedirectResponse(url=cfg["api"]["default_route"])
        return response

    # Update OpenAPI schema generator function
    if kwargs.pop("update_openapi_schema", False):
        update_openapi_kwargs = kwargs.pop("update_openapi_kwargs", {})
        get_open_api_kwargs = update_openapi_kwargs.pop("get_openapi_kwargs", {})
        get_open_api_kwargs.update({"routes": fast_api.routes})
        custom_openapi_fn = create_custom_openapi_function(
            get_open_api_kwargs,
            fix_schema=update_openapi_kwargs.get("fix_schema", False),
            auth_service=update_openapi_kwargs.get("auth_service"),
        )
        setattr(fast_api, "openapi", custom_openapi_fn)

    return fast_api



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later
from typing import ClassVar
from uuid import UUID

from pydantic import Field, field_serializer

from gen_epix.casedb.domain import enum
from gen_epix.casedb.domain.model.case.persistable import (
    GeneticDistanceProtocol,
    TreeAlgorithm,
)
from gen_epix.commondb.domain.model.base import Model
from gen_epix.commondb.util import copy_model_field
from gen_epix.fastapp import Entity
from gen_epix.fastapp.domain import Entity
from gen_epix.fastapp.domain.util import create_keys, create_links
from gen_epix.seqdb.domain.model import AssemblyProtocol as SeqdbAssemblyProtocol
from gen_epix.seqdb.domain.model import File as SeqdbFile
from gen_epix.seqdb.domain.model import LibraryPrepProtocol as SeqdbLibraryPrepProtocol
from gen_epix.seqdb.domain.model import RawSeq as SeqdbRawSeq
from gen_epix.seqdb.domain.model import ReadSet as SeqdbReadSet
from gen_epix.seqdb.domain.model import Sample as SeqdbSample
from gen_epix.seqdb.domain.model import Seq as SeqdbSeq


class GeneticSequence(Model):
    """
    A genetic sequence. Temporary implementation.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="genetic_sequences",
    )
    nucleotide_sequence: str | None = Field(
        default=None, description="The nucleotide sequence"
    )
    distances: dict[UUID, float] | None = Field(
        default=None, description="The distances to other sequences"
    )

    @field_serializer("distances", mode="plain")
    def _serialize_distances(
        self, value: dict[UUID, float] | None
    ) -> dict[str, float] | None:
        return None if value is None else {str(x): y for x, y in value.items()}


class AlleleProfile(Model):
    """
    An allele profile. Temporary implementation.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="allele_profiles",
    )
    # TODO: add link to sequence and gene set
    allele_profile: str | None = Field(default=None, description="The allele profile")


class PhylogeneticTree(Model):
    """
    A phylogenetic tree, including a description of the leaves and how it was
    generated.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="phylogenetic_trees",
        persistable=False,
    )
    tree_algorithm_id: UUID | None = Field(
        default=None, description="The ID of the tree algorithm. FOREIGN KEY"
    )
    tree_algorithm: TreeAlgorithm = Field(
        default=None, description="The tree algorithm"
    )
    tree_algorithm_code: enum.TreeAlgorithmType = Field(
        description="The tree algorithm"
    )
    genetic_distance_protocol_id: UUID | None = Field(
        default=None, description="The ID of the genetic distance protocol. FOREIGN KEY"
    )
    genetic_distance_protocol: GeneticDistanceProtocol = Field(
        default=None, description="The genetic distance protocol"
    )
    leaf_ids: list[UUID] | None = Field(
        default=None,
        description="The list of unique identifiers of the leaves of the phylogenetic tree.",
    )
    sequence_ids: list[UUID] | None = Field(
        default=None,
        description="The list of unique identifiers of the sequence of each leaf of the phylogenetic tree.",
    )
    newick_repr: str = Field(
        description="The Newick representation of the phylogenetic tree."
    )


class LibraryPrepProtocol(SeqdbLibraryPrepProtocol):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="library_prep_protocols",
        table_name="library_prep_protocol",
        persistable=False,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )


class File(SeqdbFile):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="files",
        table_name="file",
        persistable=False,
    )


class ReadSet(SeqdbReadSet):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="read_sets",
        table_name="read_set",
        persistable=False,
        keys=create_keys({1: "code"}),
        links=create_links(
            {
                1: (
                    "library_prep_protocol_id",
                    LibraryPrepProtocol,
                    "library_prep_protocol",
                ),
            }
        ),
    )
    library_prep_protocol: LibraryPrepProtocol | None = copy_model_field(SeqdbReadSet, "library_prep_protocol")

class AssemblyProtocol(SeqdbAssemblyProtocol):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="assembly_protocols",
        table_name="assembly_protocol",
        persistable=False,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )


class RawSeq(SeqdbRawSeq):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="raw_seqs",
        table_name="raw_seq",
        persistable=False,
    )


class Sample(SeqdbSample):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="samples",
        table_name="sample",
        persistable=False,
        keys=create_keys({1: "code"}),
    )


class Seq(SeqdbSeq):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="seqs",
        table_name="seq",
        persistable=False,
        keys=create_keys(
            {
                1: "code",
            }
        ),
        links=create_links(
            {
                1: ("sample_id", Sample, "sample"),
                2: ("read_set_id", ReadSet, "read_set"),
                3: ("read_set2_id", ReadSet, "read_set2"),
                4: ("assembly_protocol_id", AssemblyProtocol, "assembly_protocol"),
                5: ("raw_seq_id", RawSeq, "raw_seq"),
            }
        ),
    )
    sample: Sample | None = copy_model_field(SeqdbSeq, "sample")
    read_set: ReadSet | None = copy_model_field(SeqdbSeq, "read_set")
    read_set2: ReadSet | None = copy_model_field(SeqdbSeq, "read_set2")
    assembly_protocol: AssemblyProtocol | None = copy_model_field(SeqdbSeq, "assembly_protocol")
    raw_seq: RawSeq | None = copy_model_field(SeqdbSeq, "raw_seq")



from uuid import UUID

from gen_epix.casedb.domain import command, enum, model
from gen_epix.casedb.domain.policy.abac import BaseCaseAbacPolicy
from gen_epix.casedb.services.case.base import BaseCaseService
from gen_epix.commondb.util import map_paired_elements
from gen_epix.fastapp.enum import CrudOperation
from gen_epix.filter.base import Filter
from gen_epix.filter.uuid_set import UuidSetFilter


def case_service_retrieve_case_type_stats(
    self: BaseCaseService,
    cmd: command.RetrieveCaseTypeStatsCommand,
) -> list[model.CaseTypeStat]:
    user, repository = self._get_user_and_repository(cmd)
    assert isinstance(user, model.User) and user.id is not None
    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
    assert case_abac is not None
    case_type_ids = cmd.case_type_ids
    with repository.uow() as uow:
        cases: list[model.Case] = (
            self._retrieve_cases_with_content_right(  # type:ignore[attr-defined]
                uow,
                user.id,
                case_abac,
                # user_case_access,
                enum.CaseRight.READ_CASE,
                datetime_range_filter=cmd.datetime_range_filter,
                filter_content=False,
            )
        )
        if case_type_ids is not None:
            cases = [x for x in cases if x.case_type_id in case_type_ids]
        else:
            case_type_ids = {x.case_type_id for x in cases}
        # Derive stats
        empty_stat = {
            "n_cases": 0,
            "first_case_month": None,
            "last_case_month": None,
        }
        stats = {x: dict(empty_stat) for x in case_type_ids}
        for case in cases:
            case_type_id = case.case_type_id
            date_ = case.case_date
            stat = stats[case_type_id]
            if stat["n_cases"] == 0:
                stat["n_cases"] = 1
                stat["first_case_month"] = date_  # type: ignore[assignment]
                stat["last_case_month"] = date_  # type: ignore[assignment]
            else:
                stat["n_cases"] += 1  # type:ignore[operator]
                stat["first_case_month"] = min(stat["first_case_month"], date_)  # type: ignore[type-var,assignment]
                stat["last_case_month"] = max(stat["last_case_month"], date_)  # type: ignore[type-var,assignment]
        # Convert first/last date to month only
        for stat in stats.values():
            for key in ("first_case_month", "last_case_month"):
                stat[key] = stat[key].isoformat()[0:7]  # type: ignore[union-attr]
        # Get case type stats
        case_type_stats = [
            model.CaseTypeStat(case_type_id=x, **stats[x]) for x in case_type_ids  # type: ignore[arg-type]
        ]
    return case_type_stats


def case_service_retrieve_case_set_stats(
    self: BaseCaseService,
    cmd: command.RetrieveCaseSetStatsCommand,
) -> list[model.CaseSetStat]:
    user, repository = self._get_user_and_repository(cmd)
    case_set_ids = cmd.case_set_ids
    # Create filter, even if no case_set_ids are provided, to avoid unallowed read
    # all without filter
    query_filter: Filter | None = None
    if case_set_ids:
        query_filter = UuidSetFilter(
            key="case_set_id", members=cmd.case_set_ids  # type: ignore[arg-type]
        )
    with self.repository.uow() as uow:
        curr_cmd = command.CaseSetMemberCrudCommand(
            user=user,  # type: ignore[arg-type]
            operation=CrudOperation.READ_ALL,
            query_filter=query_filter,
        )
        curr_cmd._policies.extend(cmd._policies)
        case_set_members: list[model.CaseSetMember] = self.crud(curr_cmd)  # type: ignore[assignment]
        case_set_case_ids: dict[UUID, set[UUID]] = map_paired_elements(  # type: ignore[assignment]
            ((x.case_set_id, x.case_id) for x in case_set_members), as_set=True
        )
        if not case_set_ids:
            case_set_ids = list(case_set_case_ids.keys())
        # Get cases
        # @ABAC: case_set_case_ids is already filtered on cases with access, no
        # need to apply here again
        cases_: list[model.Case] = self.repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.Case,
            None,
            list(set.union(set(), *list(case_set_case_ids.values()))),
            CrudOperation.READ_SOME,
        )
        cases = {x.id: x for x in cases_}
        # Create case set stats
        case_set_stats = []
        case_dates = {x.id: x.case_date for x in cases.values()}
        all_case_ids = set(cases.keys())
        for case_set_id in case_set_ids:
            case_ids = case_set_case_ids.get(case_set_id, set()).intersection(
                all_case_ids
            )
            # TODO: calculate n_own_cases as the number of cases with a created_in data collection that is associated with the user
            n_own_cases = 0
            first_case_month = (
                min(case_dates[x] for x in case_ids).isoformat()[0:7]
                if case_ids
                else None
            )
            last_case_month = (
                max(case_dates[x] for x in case_ids).isoformat()[0:7]
                if case_ids
                else None
            )
            case_set_stats.append(
                model.CaseSetStat(
                    case_set_id=case_set_id,
                    n_cases=len(case_ids),
                    n_own_cases=n_own_cases,
                    first_case_month=first_case_month,
                    last_case_month=last_case_month,
                )
            )

    return case_set_stats



from collections.abc import Callable
from typing import Any

from pydantic import BaseModel, ConfigDict, model_validator

from gen_epix.fastapp.app import App
from gen_epix.fastapp.enum import CrudEndpointType
from gen_epix.fastapp.model import CrudCommand
from gen_epix.filter import Filter


class CrudEndpointSet(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())
    model_class: type
    create_api_model_class: type | None = None
    read_api_model_class: type | None = None
    endpoint_basename: str
    crud_command_class: type[CrudCommand]
    endpoint_types: set[CrudEndpointType]
    user_dependency: Callable | None = None
    app: App
    id_class: type
    operation_id_basename: str | None = None
    description: str | None = None
    post_returns_id: bool | None = False
    put_returns_id: bool | None = False
    delete_all_returns_id: bool | None = False
    response_model_exclude_none: bool | None = False
    query_filter_validator: Callable[[Filter], bool] | None = None

    @model_validator(mode="before")
    @classmethod
    def _validate_args(cls, data: Any) -> Any:
        if isinstance(data, dict):
            if not data.get("read_api_model_class"):
                data["read_api_model_class"] = data["model_class"]
            if not data.get("create_api_model_class"):
                data["create_api_model_class"] = data["read_api_model_class"]
            if data.get("operation_id_basename"):
                data["operation_id_basename"] = data["endpoint_basename"]
        else:
            raise NotImplementedError("Not implemented for non-dict data")
        return data



import abc
from collections.abc import Iterable

from gen_epix.fastapp import BaseService
from gen_epix.seqdb.domain import command, model
from gen_epix.seqdb.domain.enum import ServiceType


class BaseSeqService(BaseService):
    SERVICE_TYPE = ServiceType.SEQ

    def register_handlers(self) -> None:
        f = self.app.register_handler
        self.register_default_crud_handlers()
        f(command.RetrieveCompleteAlleleProfileCommand, self.retrieve_allele_profile)
        f(command.RetrieveCompleteSeqCommand, self.retrieve_seq)
        f(command.RetrieveCompleteSampleCommand, self.retrieve_sample)
        f(
            command.RetrieveCompleteSnpProfileCommand,
            self.retrieve_snp_profile,
        )
        f(
            command.RetrievePhylogeneticTreeCommand,
            self.retrieve_phylogenetic_tree,
        )
        f(
            command.RetrieveMultipleAlignmentCommand,
            self.retrieve_multiple_alignment,
        )
        f(
            command.RetrieveSeqFastaCommand,
            self.retrieve_seq_fasta,
        )

    @abc.abstractmethod
    def retrieve_allele_profile(
        self,
        cmd: command.RetrieveCompleteAlleleProfileCommand,
    ) -> model.CompleteAlleleProfile | list[model.CompleteAlleleProfile]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_seq(
        self,
        cmd: command.RetrieveCompleteSeqCommand,
    ) -> model.CompleteSeq | list[model.CompleteSeq]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_sample(
        self,
        cmd: command.RetrieveCompleteSampleCommand,
    ) -> model.CompleteSample | list[model.CompleteSample]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_snp_profile(
        self,
        cmd: command.RetrieveCompleteSnpProfileCommand,
    ) -> model.CompleteSnpProfile | list[model.CompleteSnpProfile]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_phylogenetic_tree(
        self, cmd: command.RetrievePhylogeneticTreeCommand
    ) -> model.PhylogeneticTree | None:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_multiple_alignment(
        self,
        cmd: command.RetrieveMultipleAlignmentCommand,
    ) -> model.MultipleAlignment | list[model.MultipleAlignment]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_seq_fasta(self, cmd: command.RetrieveSeqFastaCommand) -> Iterable[str]:
        raise NotImplementedError()



# pylint: disable=useless-import-alias
from gen_epix.casedb.domain.repository.case import (
    BaseCaseRepository as BaseCaseRepository,
)
from gen_epix.casedb.domain.repository.geo import BaseGeoRepository as BaseGeoRepository
from gen_epix.casedb.domain.repository.ontology import (
    BaseOntologyRepository as BaseOntologyRepository,
)
from gen_epix.casedb.domain.repository.subject import (
    BaseSubjectRepository as BaseSubjectRepository,
)
from gen_epix.commondb.domain.repository import BaseAbacRepository as BaseAbacRepository
from gen_epix.commondb.domain.repository import (
    BaseOrganizationRepository as BaseOrganizationRepository,
)
from gen_epix.commondb.domain.repository import (
    BaseSystemRepository as BaseSystemRepository,
)



from __future__ import annotations


from gen_epix.fastapp.model import Command
from gen_epix.seqdb.domain.service import BaseAbacService


class AbacService(BaseAbacService):
    CACHE_INVALIDATION_COMMANDS: tuple[type[Command], ...] = tuple()
